"""
This module defines the VPN widget, which contains all the VPN functionality
that is shown to the user.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from concurrent.futures import Future
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional
import time

from gi.repository import GObject, GLib

from proton.vpn import logging

from proton.vpn.connection.states import State
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.vpn.quick_connect_widget import QuickConnectWidget
from proton.vpn.app.gtk.widgets.vpn.search_results import SearchResults
from proton.vpn.app.gtk.widgets.vpn.search_entry import SearchEntry
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.serverlist import ServerListWidget
from proton.vpn.app.gtk.widgets.vpn.connection_status_widget import VPNConnectionStatusWidget
from proton.vpn.app.gtk.widgets.main.notifications import Notifications
from proton.vpn.session.servers import ServerList

if TYPE_CHECKING:
    from proton.vpn.app.gtk.app import MainWindow

logger = logging.getLogger(__name__)

# The feature flag that enables the lazy loading serverlist UI
LINUX_DEFERRED_UI = "LinuxDeferredUI"


@dataclass
class VPNWidgetState:
    """
    Holds the state of the VPNWidget. This state is reset after login/logout.

    Attributes:
        is_widget_ready: flag set to True once the widget has been initialized.
        user_tier: tier of the logged-in user.
        load_start_time: timestamp set when the widget starts loading.
    """
    is_widget_ready: bool = False
    user_tier: Optional[int] = None
    load_start_time: Optional[float] = None


# pylint: disable=too-many-instance-attributes
class VPNWidget(Gtk.Box):
    """Exposes the ProtonVPN product functionality to the user."""

    def __init__(
        self, controller: Controller,
        main_window: "MainWindow",
        notifications=Notifications
    ):
        super().__init__(spacing=10)

        self.set_name("vpn-widget")
        self._state = VPNWidgetState()
        self._state.load_start_time = time.time()
        self._controller = controller

        self.connection_status_widget = VPNConnectionStatusWidget(
            controller, notifications
        )
        self.append(self.connection_status_widget)

        self._connected_signals: list[tuple[int, Gtk.Widget]] = []

        self.quick_connect_widget = QuickConnectWidget(self._controller)
        self.append(self.quick_connect_widget)

        self.search_widget = SearchEntry()
        self.server_list_widget = ServerListWidget(self._controller, self.search_widget)
        self.append(self.server_list_widget)
        self._connected_signals.append((
            safe_signal_connect(
                self.server_list_widget, "ui-updated", self._on_server_list_updated
            ),
            self.server_list_widget
        ))
        main_window.add_keyboard_shortcut(
            target_widget=self.search_widget,
            target_signal="request_focus",
            shortcut="<Control>f"
        )
        self.search_results_widget = SearchResults(self._controller)
        revealer = Gtk.Revealer()
        revealer.set_child(self.search_results_widget)
        self.search_results_widget.set_revealer(revealer)

        self._connected_signals.append((
            safe_signal_connect(
                self.search_widget,
                "search-changed",
                self.search_results_widget.on_search_changed,
            ),
            self.search_widget
        ))
        self._connected_signals.append((
            safe_signal_connect(
                self.search_results_widget,
                "result-chosen",
                self.server_list_widget.focus_on_entry
            ),
            self.search_results_widget
        ))
        self._connected_signals.append((
            safe_signal_connect(
                self.search_results_widget,
                "result-chosen",
                self._reset_search_on_result_chosen
            ),
            self.search_results_widget
        ))
        self.insert_child_after(self.search_widget, self.quick_connect_widget)
        self.insert_child_after(revealer, self.search_widget)

        self._state_subscribers = [
            self.connection_status_widget,
            self.quick_connect_widget,
        ]

        signal_id = safe_signal_connect(
            self,
            "connection-state-changed",
            self._broadcast_connection_state
        )
        self._connected_signals.append((signal_id, self))

        self.set_orientation(Gtk.Orientation.VERTICAL)

        safe_signal_connect(self, "unrealize", self._on_unrealize)

    def _reset_search_on_result_chosen(self, *_) -> None:
        self.search_widget.reset()

    def _broadcast_connection_state(self, _, state):
        for widget in self._state_subscribers:
            widget.connection_status_update(state)

    @GObject.Signal
    def vpn_widget_ready(self):
        """Signal emitted when all resources were loaded and widget is ready."""

    @GObject.Signal(name="connection-state-changed", arg_types=(object,))
    def connection_state_changed(self, state: State):
        """Signal emitted whenever the VPN connection state changes."""

    @property
    def user_tier(self) -> int:
        """Returns the tier of the user currently logged in."""
        return self._state.user_tier

    def _on_unrealize(self, _widget):
        self.unload()

    def status_update(self, connection_state: State):
        """This method is called whenever the VPN connection status changes."""
        logger.debug(
            f"VPN widget received connection status update: "
            f"{type(connection_state).__name__}."
        )

        GLib.idle_add(self.emit, "connection-state-changed", connection_state)

    def _on_refresher_enabled(
            self,
            future: Future
    ):
        future.result()
        self.display(self._controller.user_tier, self._controller.server_list)

    def load(self):
        """
        Starts loading the widget.

        The call to this method triggers networks calls to Proton's REST API
        to download the required data to display the widget. Once the required
        data has been downloaded, the widget will be automatically displayed.
        """
        self._state.load_start_time = time.time()
        self._controller.enable_refresher(self._on_refresher_enabled)

    def display(self, user_tier: int, server_list: ServerList):
        """Displays the widget once all necessary data from API has been acquired."""
        self._state.user_tier = user_tier

        # The VPN widget subscribes to connection status updates, and then
        # passes on these connection status updates to child widgets
        self._controller.register_connection_status_subscriber(self)
        self._controller.reconnector.enable()

        # Apply the correct connection state immediately so there's no flash of
        # "Unprotected" when the app starts already connected.
        self.status_update(self._controller.current_connection_status)

        self.server_list_widget.display(user_tier=user_tier, server_list=server_list)

    def _on_server_list_updated(self, *_):
        if not self._state.is_widget_ready:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            # Only update the status at this point as widgets are already generated
            self.status_update(self._controller.current_connection_status)
            self._state.is_widget_ready = True  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            self.emit("vpn-widget-ready")
            logger.info(
                f"VPN widget is ready "
                f"(load time: {time.time()-self._state.load_start_time:.2f} seconds)",
                category="app", subcategory="vpn", event="widget_ready"
            )

    def unload(self):
        """Unloads the widget and resets its state."""
        for signal_id, widget in self._connected_signals:
            widget.disconnect(signal_id)
        self._connected_signals.clear()
        self._controller.disconnect()

        self._controller.unregister_connection_status_subscriber(self)
        self._controller.reconnector.disable()
        self._controller.disable_refresher()

        for widget in [
            self.connection_status_widget,
            self.quick_connect_widget, self.server_list_widget
        ]:
            widget.set_visible(False)

        # Reset widget state
        self._state = VPNWidgetState()



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Callable

from proton.vpn import logging

from proton.vpn.core.settings import Settings

logger = logging.getLogger(__name__)


class SettingsWatchers:
    """
    A class to manage watchers for settings changes.
    It allows adding, removing, and notifying watchers when settings change.
    Each watcher is a callable that takes the updated settings as an argument.
    """
    def __init__(self):
        self.watchers = []

    def add(self, watcher: Callable[[Settings], None]):
        """Adds a new watcher to the list."""
        self.watchers.append(watcher)

    def remove(self, watcher: Callable[[Settings], None]):
        """Removes a watcher from the list."""
        if watcher in self.watchers:
            self.watchers.remove(watcher)

    def notify(self, settings: Settings):
        """Notifies all watchers with the new settings."""
        for watcher in self.watchers:
            watcher(settings)

    def clear(self):
        """Clears all watchers."""
        self.watchers.clear()



"""
This module defines the login widget, used to authenticate the user.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import TYPE_CHECKING, Optional

from gi.repository import GObject

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.main.confirmation_dialog import ConfirmationDialog

if TYPE_CHECKING:
    from proton.vpn.app.gtk.app import MainWindow


class KillSwitchLabel(Gtk.Label):
    """Label objet that already contains some styling and pre-configurations"""
    LABEL_TEXT = "Kill switch is blocking any outgoing connections."

    def __init__(self):
        super().__init__(label=KillSwitchLabel.LABEL_TEXT)
        self.set_wrap(True)
        # set_max_width_chars is required for set_line_wrap to have effect.
        self.set_max_width_chars(20)
        self.set_justify(Gtk.Justification.LEFT)


class DisableKillSwitchButton(Gtk.Button):
    """Custom button that already has styling added to it."""
    BUTTON_LABEL = "Disable"

    def __init__(self):
        super().__init__(label=DisableKillSwitchButton.BUTTON_LABEL)
        self.add_css_class("secondary")
        self.add_css_class("spaced")


class DisableKillSwitchWidget(Gtk.Revealer):
    """This is a revealer that displays a short informational message
    and a button it disable the kill switch.
    """
    DIALOG_TITLE = "Kill Switch Enabled"
    DIALOG_MESSAGE = "Permanent Kill Switch is blocking any outgoing connections "\
        "and preventing your IP to be exposed.\n\n"\
        "Do you want to disable Kill Switch ?"

    def __init__(
        self, main_window: "MainWindow",
        killswitch_label: Optional[KillSwitchLabel] = None,
        killswitch_button: Optional[DisableKillSwitchButton] = None,
    ):
        super().__init__()
        self.set_name("login-kill-switch-revealer")

        self._main_window = main_window

        self.killswitch_label = killswitch_label or KillSwitchLabel()
        self.disable_killswitch_button = killswitch_button or DisableKillSwitchButton()

        container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        container.append(self.killswitch_label)
        container.append(self.disable_killswitch_button)

        self.set_child(container)

        safe_signal_connect(
            self.disable_killswitch_button,
            "clicked",
            self._on_button_click
        )

    def _on_button_click(self, _: Gtk.Button):
        dialog = ConfirmationDialog(
            DisableKillSwitchWidget.DIALOG_MESSAGE,
            DisableKillSwitchWidget.DIALOG_TITLE
        )
        dialog.set_transient_for(self._main_window)
        dialog.set_modal(True)

        safe_signal_connect(
            dialog,
            "response",
            self._on_dialog_response
        )
        dialog.present()

    def _on_dialog_response(self, dialog: ConfirmationDialog, response_id: int):
        dialog.destroy()
        if response_id == Gtk.ResponseType.YES:
            self.emit("disable-killswitch")

    @GObject.Signal
    def disable_killswitch(self):
        """Signal emitted when the user confirms to disable kill switch."""



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from pathlib import Path

from gi.repository import Gdk, Gtk
from proton.vpn.app.gtk.assets import icons
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect


class PasswordEntry(Gtk.Entry):
    """
    Entry used to introduce the password in the login form.

    On top of the inherited functionality from Gtk.Entry, an icon is shown
    inside the text entry to show or hide the password.

    By default, the text (password) introduced in the entry is not show.
    Therefore, the icon to be able to show the text is displayed. Once this
    icon is pressed, the text is revealed and the icon to hide the password
    is shown instead.
    """
    def __init__(self):
        super().__init__()
        self.set_input_purpose(Gtk.InputPurpose.PASSWORD)
        self.set_visibility(False)
        # Load icon to hide the password.
        eye_dirpath = Path("eye")
        hide_fp = eye_dirpath / "hide.svg"
        self._hide_pixbuff = icons.get(
            hide_fp,
            width=17,
            height=17,
            preserve_aspect_ratio=True
        )
        # Load icon to show the password.
        show_fp = eye_dirpath / "show.svg"
        self._show_pixbuff = icons.get(
            show_fp,
            width=17,
            height=17,
            preserve_aspect_ratio=True
        )
        # By default, the password is not shown. Therefore, the icon to
        # be able to show the password is shown.
        show_texture = Gdk.Texture.new_for_pixbuf(self._show_pixbuff)
        self.set_icon_from_gicon(
            Gtk.EntryIconPosition.SECONDARY,
            show_texture
        )
        self.set_icon_activatable(
            Gtk.EntryIconPosition.SECONDARY,
            True
        )
        safe_signal_connect(
            self, "icon-press", self._on_change_password_visibility_icon_press
        )

    def _on_change_password_visibility_icon_press(self, gtk_entry_object, _icon_position):
        """Changes password visibility, updating accordingly the icon."""
        is_text_visible = gtk_entry_object.get_visibility()
        gtk_entry_object.set_visibility(not is_text_visible)
        pixbuf = self._show_pixbuff if is_text_visible else self._hide_pixbuff
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_icon_from_gicon(
            Gtk.EntryIconPosition.SECONDARY,
            texture
        )



"""
This module defines the two factor authentication stack, used to contain
the different 2FA methods.


Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from concurrent.futures import Future
from typing import Union, Optional
from gi.repository import GObject, Gtk

from proton.vpn import logging
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.util import connect_once
from proton.vpn.app.gtk.utils.glib import add_done_callback
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.main.notifications import Notifications
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget
from proton.vpn.app.gtk.widgets.login.two_factor_auth.authenticator_app_form \
    import AuthenticatorAppForm
from proton.vpn.app.gtk.widgets.login.two_factor_auth.security_key_form \
    import SecurityKeyForm

logger = logging.getLogger(__name__)


class TwoFactorAuthStack(Gtk.Stack):
    """Stack used to display the 2FA methods."""
    AUTHENTICATOR_APP_FORM_TITLE = "Authenticator app"
    SECURITY_KEY_FORM_TITLE = "Security key"

    def __init__(  # pylint: disable=too-many-arguments
        self,
        controller: Controller,
        notifications: Notifications,
        overlay_widget: OverlayWidget,
        authenticator_app_form: Optional[AuthenticatorAppForm] = None,
        security_key_form: Optional[SecurityKeyForm] = None
    ):
        super().__init__()

        self.set_name("two-factor-auth-stack")
        self._controller = controller
        self._notifications = notifications
        self._overlay_widget = overlay_widget
        self.active_widget: Optional[Union[AuthenticatorAppForm, SecurityKeyForm]] = None
        self.set_transition_type(Gtk.StackTransitionType.SLIDE_LEFT_RIGHT)
        self._pending_logout_future = None

        # SecurityKeyForm
        self.security_key_form = security_key_form or SecurityKeyForm(
            controller, notifications, overlay_widget
        )
        safe_signal_connect(
            self.security_key_form,
            "two-factor-auth-successful",
            self._on_two_factor_auth_successful
        )
        safe_signal_connect(
            self.security_key_form,
            "two-factor-auth-cancelled",
            self._on_two_factor_auth_cancelled
        )

        # AuthenticatorAppForm
        self.authenticator_app_form = authenticator_app_form or AuthenticatorAppForm(
            controller, notifications, overlay_widget
        )
        safe_signal_connect(
            self.authenticator_app_form,
            "two-factor-auth-successful",
            self._on_two_factor_auth_successful
        )
        safe_signal_connect(
            self.authenticator_app_form,
            "two-factor-auth-cancelled",
            self._on_two_factor_auth_cancelled
        )

    def display_widget(self, widget: Union[AuthenticatorAppForm, SecurityKeyForm]):
        """
        Displays the specified form to the user.
        """
        if widget is not self.get_child_by_name("authenticator_app_form") and \
           widget is not self.get_child_by_name("security_key_form"):
            raise ValueError("Invalid widget to display in TwoFactorAuthStack")

        self.active_widget = widget
        self.set_visible_child(widget)
        widget.reset()

    def reset(self):
        """Resets the widget to its initial state."""
        self._notifications.hide_message()

        # Remove all children first
        if self.get_child_by_name("security_key_form"):
            self.remove(self.security_key_form)
        if self.get_child_by_name("authenticator_app_form"):
            self.remove(self.authenticator_app_form)

        if self._controller.fido2_available:
            self.add_titled(
                self.security_key_form,
                "security_key_form",
                self.SECURITY_KEY_FORM_TITLE
            )
            self.add_titled(
                self.authenticator_app_form,
                "authenticator_app_form",
                self.AUTHENTICATOR_APP_FORM_TITLE
            )
            self.display_widget(self.security_key_form)
        else:
            self.add_titled(
                self.authenticator_app_form,
                "authenticator_app_form",
                self.AUTHENTICATOR_APP_FORM_TITLE
            )
            self.display_widget(self.authenticator_app_form)

    def _on_two_factor_auth_successful(self, _):
        self.emit("two-factor-auth-successful")

    def _on_two_factor_auth_cancelled(self, _):
        logger.info(
            "2FA cancelled by user, signing out...",
            category="UI", subcategory="LOGIN-2FA", event="CLICK"
        )
        self._overlay_widget.show_message("Signing out...")
        future = self._controller.logout()
        add_done_callback(future, self._on_logout)

    def _on_logout(self, logout_future: Future):
        self._pending_logout_future = logout_future
        connect_once(
            self._overlay_widget,
            "hide",
            self._on_overlay_hidden_after_logout,
        )
        self._overlay_widget.hide()

    def _on_overlay_hidden_after_logout(self, _overlay_widget: OverlayWidget):
        self._pending_logout_future.result()
        self._pending_logout_future = None
        self.emit("two-factor-auth-cancelled")

    @GObject.Signal
    def two_factor_auth_successful(self):
        """Signal emitted after a successful 2FA."""

    @GObject.Signal
    def two_factor_auth_cancelled(self):
        """Signal emitted after 2FA was cancelled by the user."""



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
import concurrent
import functools
import inspect
from concurrent.futures import ThreadPoolExecutor
from threading import Thread
from typing import Optional, Coroutine, Callable, Union

from proton.vpn import logging

logger = logging.getLogger(__name__)


class AsyncExecutor:
    """
    Allows non-asyncio code to execute both coroutine functions and regular (blocking) functions
    in a non-blocking manner.

    Usage:

    .. code-block:: python
        import asyncio
        import time

        async def async_func(blocking_time: int) -> str:
            await asyncio.sleep(blocking_time)
            return "async func done"

        def regular_func(blocking_time: int) -> str:
            time.sleep(blocking_time)
            return "regular func done"

        with AsyncExecutor() as ce:
            future1 = ce.submit(async_func, blocking_time=0)
            future2 = ce.submit(regular_func, blocking_time=0)
            assert future1.result() == "async func done"
            assert future2.result() == "regular func done"

    """

    def __init__(
            self, loop: Optional[asyncio.AbstractEventLoop] = None,
            executor: Optional[ThreadPoolExecutor] = None
    ):
        self._thread: Optional[Thread] = None
        self._executor = executor or ThreadPoolExecutor()
        self._loop = loop or asyncio.new_event_loop()

    def start(self):
        """
        Starts the async executor.

        It starts a thread that runs the asyncio loop.
        """
        if self.is_running:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            raise RuntimeError("The executor is already running.")

        self._thread = Thread(target=self._run_asyncio_loop_forever, daemon=True)
        self._thread.start()

    @property
    def is_running(self) -> bool:
        """Returns True if the async executor has already been started and False otherwise."""
        return self._thread is not None

    def _run_asyncio_loop_forever(self):
        self._executor = ThreadPoolExecutor()
        self._loop.set_default_executor(self._executor)
        asyncio.set_event_loop(self._loop)
        try:
            self._loop.run_forever()
        finally:
            # Currently recommended way of shutting down the loop:
            # https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.close
            self._loop.run_until_complete(self._loop.shutdown_asyncgens())
            self._loop.run_until_complete(self._loop.shutdown_default_executor())
            self._loop.close()

    def stop(self):
        """
        Stops the async executor.

        It schedules a call to stop the asyncio loop and waits for the thread
        running it to stop.
        """
        if not self.is_running:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            logger.warning("The executor has already been stopped.")
            return

        self._loop.call_soon_threadsafe(self._loop.stop)
        self._thread.join()
        self._thread = None

    # pylint: disable=invalid-name
    def submit(self, fn: Union[Coroutine, Callable], *args, **kwargs) -> concurrent.futures.Future:
        """
        Submits a coroutine function or a callable to be run on the async executor
        in a thread-safe manner and non-blocking manner.

        :returns: a Future that can be waited for in a non-asyncio manner (or not).
        """
        if inspect.iscoroutinefunction(fn):
            coroutine = fn(*args, **kwargs)
            return asyncio.run_coroutine_threadsafe(coroutine, self._loop)

        return self._executor.submit(fn, *args, **kwargs)

    async def _blocking_function_to_coroutine(self, fn, *args, **kwargs):
        fn_wrapper = functools.partial(fn, *args, **kwargs)
        return await self._loop.run_in_executor(executor=None, func=fn_wrapper)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



"""
Utility functions for server list widgets.


Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Any, Callable, List, Type, TypeVar

from gi.repository import GLib

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.session.servers import ServerList

GtkWidget = TypeVar("GtkWidget", bound=Gtk.Widget)


def make_connect_callback(
    controller: Controller, servers: list, user_tier: int
) -> Callable[[], None]:
    """Returns a callback that connects to the fastest available server in the given list."""
    def on_connect():
        fastest = ServerList.get_fastest_server(
            ServerList.get_available_servers(servers=servers, user_tier=user_tier)
        )
        future = controller.connect_to_server(fastest.name)
        future.add_done_callback(lambda f: GLib.idle_add(f.result))
    return on_connect


def sync_rows_with_model_items(
    model_items: List[Any],
    rows: List[GtkWidget],
    container: Gtk.Box,
    row_factory: Type[GtkWidget],
    display_func: Callable[[GtkWidget, Any], None]
):
    """Synchronizes a list of row widgets with model items.

    This utility function handles the common pattern of:
    - Creating new rows when there are more model items than rows
    - Removing rows when there are more rows than model items
    - Updating existing rows with their corresponding model items

    ``rows`` is the source of truth and is mutated in place.
    Row widgets should connect their reset() method to the "unrealize"
    signal for automatic cleanup when removed.

    Args:
        model_items: List of model objects to display
        rows: Mutable list of row widgets owned by the caller
        container: Container widget to add/remove rows from
        row_factory: Callable that creates a new row widget (no arguments)
        display_func: Callable(row, model_item) that updates a row with a model item
    """
    for i, model_item in enumerate(model_items):
        if i >= len(rows):
            row = row_factory()
            rows.append(row)
            container.append(row)
        display_func(rows[i], model_item)

    while len(rows) > len(model_items):
        row = rows.pop()
        container.remove(row)


def get_children(widget: Gtk.Widget) -> List[Gtk.Widget]:
    """Returns the children of a widget."""
    children = []
    child = widget.get_first_child()
    while child:
        children.append(child)
        child = child.get_next_sibling()
    return children



"""
This module defines the Loading widget. This widget is responsible for displaying
the loading screen.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional, Union
from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect


class Spinner(Gtk.Spinner):
    """Spinner with some default configurations.
    Upon being shown it automatically starts spinning.
    """
    def __init__(self, size: int = 50):
        super().__init__()
        self.set_property("height-request", size)
        safe_signal_connect(self, "realize", self._on_realize)

    def _on_realize(self, _: Gtk.Widget):
        """Starts spinning when the widget is realized."""
        self.start()


class BaseLoadingContainerWidget(Gtk.Box):
    """Used mainly to standardize and styling, to reduce boilerplate code.
    """
    def __init__(self, orientation: Gtk.Orientation = Gtk.Orientation.VERTICAL):
        super().__init__(orientation=orientation)
        self.set_spacing(25)
        self.set_valign(Gtk.Align.CENTER)
        self.set_vexpand(True)


class DefaultLoadingWidget(BaseLoadingContainerWidget):
    """Helper class to be used when only a label is needed
    to be displayed with a spinner."""
    def __init__(self, label: str):
        super().__init__()
        self._label = Gtk.Label.new(label)
        self._label.set_wrap(True)
        self._label.set_max_width_chars(1)
        self._label.set_hexpand(True)
        self._label.set_justify(Gtk.Justification.CENTER)
        self._label.add_css_class("default-loading-widget-label")
        self._label.set_valign(Gtk.Align.CENTER)
        self._label.set_vexpand(True)
        self._spinner = Spinner()

        self.append(self._label)
        self.append(self._spinner)

    def get_label(self) -> str:
        """Returns the label of the object"""
        return self._label.get_label()


class LoadingConnectionWidget(BaseLoadingContainerWidget):
    """When establishing connections, this widget is used to display status,
    hide the main vpn widget and display a cancel connection button.
    """
    def __init__(
        self, label: str,
        cancel_button: Gtk.Button,
        display_loading_status: Optional[Gtk.Widget] = None
    ):
        super().__init__()

        self._label = Gtk.Label.new(label)
        self._cancel_button = cancel_button
        self._cancel_button.add_css_class("danger")
        self._cancel_button.set_halign(Gtk.Align.CENTER)

        self._display_loading_status: Union[Spinner, Gtk.Widget]
        if not display_loading_status:
            self._display_loading_status = Spinner()
            self._display_loading_status.start()
        else:
            self._display_loading_status = display_loading_status

        self.append(self._label)
        self.append(self._display_loading_status)
        self.append(self._cancel_button)

    def get_label(self) -> str:
        """Returns the label shown while the connection is being established."""
        return self._label.get_label()

    def set_label(self, label: str):
        """Sets the label shown while the connection is being established."""
        return self._label.set_label(label)


class OverlayWidget(Gtk.Box):
    """Loading widget responsible for displaying loading status
    to the user."""

    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self._centered_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=20)

        self._centered_container.set_valign(Gtk.Align.CENTER)

        self.append(self._centered_container)
        # Adding the background class (which is a GTK class) gives the default
        # background color to this widget. This is needed as otherwise the widget
        # background is transparent, but the intended use of this widget is to
        # hide other widgets while an action is ongoing.
        self.add_css_class("background")
        self.set_visible(False)

    def show(self, widget: Gtk.Widget):  # pylint: disable=arguments-differ
        """Shows the loading screen to the user."""
        self._remove_children_if_any()
        self._centered_container.append(widget)
        super().set_visible(True)

    def show_message(self, message: str):
        """Shows a message using DefaultLoadingWidget"""
        self.show(DefaultLoadingWidget(message))

    def hide(self):  # pylint: disable=arguments-differ
        """Hides the loading widget from the user."""
        self._remove_children_if_any()
        super().set_visible(False)

    def _remove_children_if_any(self):
        first_child = self._centered_container.get_first_child()
        if first_child:
            self._centered_container.remove(first_child)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
import subprocess  # nosec B404 # nosemgrep: gitlab.bandit.B404
from concurrent.futures import Future
from importlib import metadata
from threading import Event
from types import TracebackType
from typing import Optional, Type, Callable, Union, Tuple, List

from gi.repository import GLib
from proton.vpn.session import ServerList

from proton.vpn import logging

from proton.vpn.connection import VPNConnection, states
from proton.vpn.core.api import ProtonVPNAPI, VPNAccount
from proton.vpn.core.session_holder import ClientTypeMetadata
from proton.vpn.core.vpnconnector import VPNConnector
from proton.vpn.core.cache_handler import CacheHandler
from proton.vpn.core.settings import Settings
from proton.vpn.session.servers import LogicalServer
from proton.vpn.session.session import \
    FeatureFlags, \
    Notifications as PullNotifications
from proton.vpn.session.u2f_interaction import UserInteraction

from proton.vpn.connection.enum import KillSwitchSetting as\
    KillSwitchSettingEnum

from proton.vpn.app.gtk.services import VPNReconnector
from proton.vpn.app.gtk.services.reconnector.network_monitor import NetworkMonitor
from proton.vpn.app.gtk.services.reconnector.session_monitor import SessionMonitor
from proton.vpn.app.gtk.services.reconnector.vpn_monitor import VPNMonitor
from proton.vpn.app.gtk.settings_watchers import SettingsWatchers
from proton.vpn.app.gtk.utils import glib
from proton.vpn.app.gtk.utils.exception_handler import ExceptionHandler
from proton.vpn.app.gtk.utils.executor import AsyncExecutor
from proton.vpn.app.gtk.widgets.headerbar.menu.bug_report_dialog import BugReportForm
from proton.vpn.session.dataclasses import NPSSurveyResponse
from proton.vpn.app.gtk.config import AppConfig, APP_CONFIG
from proton.vpn.app.gtk.conflicts import Conflicts, Conflict

from proton.session.api import Fido2Assertion

logger = logging.getLogger(__name__)

DOT = "."  # pylint: disable=invalid-name


class Controller:  # pylint: disable=too-many-public-methods, too-many-instance-attributes
    """The C in the MVC pattern."""

    @staticmethod
    def get(executor: AsyncExecutor, exception_handler: "ExceptionHandler") -> Controller:
        """Preferred method to get an instance of Controller."""
        controller = Controller(executor, exception_handler)
        executor.submit(controller.initialize_vpn_connector).result()
        return controller

    def __init__(
        self,
        executor: AsyncExecutor,
        exception_handler: ExceptionHandler,
        api: Optional[ProtonVPNAPI] = None,
        vpn_connector: Optional[VPNConnector] = None,
        vpn_reconnector: Optional[VPNReconnector] = None,
        app_config: Optional[AppConfig] = None,
        cache_handler: CacheHandler = None
    ):  # pylint: disable=too-many-arguments
        self.executor = executor

        self.exception_handler = exception_handler
        self.exception_handler.controller = self

        client_type_metadata = ClientTypeMetadata(type="gui")

        self._api = api or ProtonVPNAPI(client_type_metadata)
        self._connector = vpn_connector
        self.reconnector = vpn_reconnector

        self._app_config = app_config
        self._cache_handler = cache_handler or CacheHandler(APP_CONFIG)
        self._settings_watchers = SettingsWatchers()

    async def initialize_vpn_connector(self):
        """
        Runs the required initializations to be able to start new VPN connections.
        """
        self._connector = await self._api.get_vpn_connector()

        self.reconnector = VPNReconnector(
            vpn_data_refresher=self._api.refresher,
            vpn_connector=self._connector,
            vpn_monitor=VPNMonitor(vpn_connector=self._connector),
            network_monitor=NetworkMonitor(pool=self.executor),
            session_monitor=SessionMonitor(),
            async_executor=self.executor
        )

    def login(self, username: str, password: str) -> Future:
        """
        Logs the user in.
        :param username:
        :param password:
        :return: A Future object wrapping the result of the login API call.
        """
        return self.executor.submit(self._api.login, username, password)

    def submit_2fa_code(self, code: str) -> Future:
        """
        Submits a 2-factor authentication code for verification.
        :param code: The 2FA code.
        :return: A Future object wrapping the result of the 2FA verification.
        """
        return self.executor.submit(self._api.submit_2fa_code, code)

    def generate_2fa_fido2_assertion(
            self,
            user_interaction: UserInteraction,
            cancel_assertion: Event) -> Future:
        """
        Scans for security keys and generates a FIDO2 assertion for U2F authentication.
        :param user_interaction: object handling any required user interaction
            while generating the assertion.
        :param cancel_assertion: optional event that can be set to cancel the
        fido 2 assertion process.
        :return: A Future object wrapping the Fido2Assertion.
        """
        return self.executor.submit(
            self._api.generate_2fa_fido2_assertion,
            user_interaction,
            cancel_assertion)

    def submit_2fa_fido2(self, fido2_assertion: Fido2Assertion) -> Future:
        """
        Submits a 2-factor authentication using U2F for verification.
        :return: A Future object wrapping the result of the 2FA verification.
        """
        return self.executor.submit(self._api.submit_2fa_fido2, fido2_assertion)

    def logout(self) -> Future:
        """
        Logs the user out.
        :return: A future to be able to track the logout completion.
        """
        return self.executor.submit(self._api.logout)

    @property
    def user_logged_in(self) -> bool:
        """
        Returns whether the user is logged in or not.
        :return: True if the user is logged in. Otherwise, False.
        """
        return self._api.is_user_logged_in()

    @property
    def user_tier(self):
        """Returns user tier."""
        return self._api.user_tier

    @property
    def settings_watchers(self):
        """Returns a registry for settings changed callbacks."""
        return self._settings_watchers

    def run_startup_actions(self, _):
        """Runs any startup actions that are necessary once the app has loaded."""
        logger.info(
            "Running startup actions",
            category="app", subcategory="startup", event="startup_actions"
        )
        if (
            self.user_logged_in
            and self.get_app_configuration().connect_at_app_startup
        ):
            self.autoconnect()

    def autoconnect(self) -> Future:
        """Connects to a server from app configuration.
            This method is intended to be called at app startup.
        """
        connect_at_app_startup = self.get_app_configuration().connect_at_app_startup

        # Temporary hack for parsing. Should be improved
        if connect_at_app_startup == "FASTEST":
            return self.connect_to_fastest_server()

        return self._connect_to(connect_at_app_startup)

    def connect_from_tray(self, connect_to: str) -> Future:
        """Connect to servers from tray."""
        return self._connect_to(connect_to)

    def _connect_to(self, connect_to: str) -> Future:
        if "#" in connect_to:
            return self.connect_to_server(connect_to)

        return self.connect_to_country(connect_to)

    def connect_to_country(self, country_code: str) -> Future:
        """
        Establishes a VPN connection to the specified country.
        :param country_code: The ISO3166 code of the country to connect to.
        :return: A Future object that resolves once the connection reaches the
        "connected" state.
        """
        server = self._api.server_list.get_fastest_in_country(country_code)
        return self._connect_to_vpn(server)

    def connect_to_fastest_server(self) -> Future:
        """
        Establishes a VPN connection to the fastest server.
        :return: A Future object that resolves once the connection reaches the
        "connected" state.
        """
        # FIXME: getting the fastest server takes too long  # pylint: disable=fixme
        server = self._api.server_list.get_fastest()
        return self._connect_to_vpn(server)

    def connect_to_server(self, server_name: Optional[str] = None) -> Future:
        """
        Establishes a VPN connection.
        :param server_name: The name of the server to connect to.
        :return: A Future object that resolves once the connection reaches the
        "connected" state.
        """
        server = self._api.server_list.get_by_name(server_name)
        return self._connect_to_vpn(server)

    def _connect_to_vpn(self, server: LogicalServer) -> Future:
        vpn_server = self._connector.get_vpn_server(
            server, self._api.refresher.client_config
        )

        return self.executor.submit(
            self._connector.connect,
            vpn_server,
            protocol=self.get_settings().protocol
        )

    def disconnect(self) -> Future:
        """
        Terminates a VPN connection.
        :return: A Future object that resolves once the connection reaches the
        "disconnected" state.
        """
        return self.executor.submit(self._connector.disconnect)

    @property
    def account_name(self) -> str:
        """Returns account name."""
        return self._api.account_name

    @property
    def account_data(self) -> VPNAccount:
        """Returns account data."""
        return self._api.account_data

    @property
    def current_connection(self) -> VPNConnection:
        """Returns the current VPN connection, if it exists."""
        return self._connector.current_connection

    @property
    def current_connection_status(self) -> states.State:
        """Returns the current VPN connection status. If there is not a
        current VPN connection, then the Disconnected state is returned."""
        return self._connector.current_state

    @property
    def current_server_id(self) -> str:
        """Returns the server id of the current connection."""
        return self._connector.current_server_id

    @property
    def is_connection_active(self) -> bool:
        """
        Returns whether the current connection is active or not.

        A connection is considered active in the connecting, connected
        and disconnecting states.
        """
        return self._connector.is_connection_active  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses

    @property
    def connection_disconnected(self) -> bool:
        """Returns whether the current connection is in disconnected state or not."""
        return isinstance(self._connector.current_state, states.Disconnected)

    @property
    def server_list(self) -> ServerList:
        """Returns the current server list."""
        return self._api.refresher.server_list

    @property
    def feature_flags(self) -> FeatureFlags:
        """Returns object which specifies which features are to be enabled or not."""
        return self._api.refresher.feature_flags

    @property
    def notifications(self) -> PullNotifications:
        """Returns cached VPN pull notifications."""
        return self._api.refresher.notifications

    def set_notification_seen(self, notification_id: str):
        """Marks a notification as seen and persists the change to disk."""
        self._api.set_notification_seen(notification_id)

    def submit_bug_report(self, bug_report: BugReportForm) -> Future:
        """Submits an issue report.
        :return: A Future object wrapping the result of the API."""
        return self.executor.submit(
            self._api.submit_bug_report,
            bug_report
        )

    def submit_nps_survey_response(self, nps_response: NPSSurveyResponse) -> Future:
        """Submits an NPS survey response.
        :return: A Future object wrapping the result of the API."""
        return self.executor.submit(
            self._api.submit_nps_response,
            nps_response
        )

    def register_connection_status_subscriber(self, subscriber):
        """
        Registers a new subscriber to connection status updates.
        :param subscriber: The subscriber to be registered.
        """
        self._connector.register(subscriber)

    def unregister_connection_status_subscriber(self, subscriber):
        """
        Unregisters an existing subscriber from connection status updates.
        :param subscriber: The subscriber to be unregistered.
        """
        self._connector.unregister(subscriber)

    @property
    def vpn_connector(self) -> VPNConnector:
        """Returns the VPN connector"""
        return self._connector

    def disable_killswitch(self) -> Future:
        """Disables the kill switch and stores the change to file."""
        settings = self.get_settings()
        settings.killswitch = KillSwitchSettingEnum.OFF
        return self.save_settings(settings)

    def get_app_configuration(self) -> AppConfig:
        """Return object with app specific configurations."""
        if self._app_config is not None:
            return self._app_config

        app_config = self._cache_handler.load()

        if app_config is None:
            self._app_config = AppConfig.default()
        else:
            self._app_config = AppConfig.from_dict(app_config)

        return self._app_config

    def save_app_configuration(self, new_value: AppConfig):
        """Save object with app specific configurations to disk."""
        self._app_config = new_value
        self._cache_handler.save(self._app_config.to_dict())

    @property
    def app_version(self) -> str:
        """Returns the current app version."""
        return metadata.version("proton-vpn-gtk-app")

    def get_settings(self) -> Settings:
        """Returns general settings."""

        return self.executor.submit(
            self._api.load_settings
        ).result()

    def save_settings(self, settings: Settings, bubble_up_errors=True) -> Future:
        """
        Saves current settings to disk and updates the wireguard certificate
        if necessary.
        """
        async def save(settings):
            await self._api.save_settings(settings)

        future = self.executor.submit(
            save,
            settings
        )

        future.add_done_callback(
            lambda f: GLib.idle_add(self._settings_watchers.notify, settings)
        )

        if bubble_up_errors:
            glib.bubble_up_errors(future)

        return future

    def resolve_settings_type(self, setting_path_name: str) -> Tuple[str,
                                                                     str,
                                                                     Callable,
                                                                     Callable]:
        """
        Resolves the type of the setting based on the setting path name.
        :param setting_path_name: The path name of the setting.
        :return: A tuple containing the setting type, setting attributes,
        and the methods to get and save the settings.
        """
        setting_type, setting_attrs = setting_path_name.split(DOT, maxsplit=1)
        if setting_type == "settings":
            return (
                setting_type,
                setting_attrs,
                self.get_settings,
                self.save_settings
            )
        if setting_type == "app_configuration":
            return (
                setting_type,
                setting_attrs,
                self.get_app_configuration,
                self.save_app_configuration
            )

        raise ValueError(
            f"Unknown setting type: {setting_type}. "
            "Expected 'settings' or 'app_config'."
        )

    def setting_attr_has_conflict(self, setting_path_name: str,
                                  new_value: object) -> Optional[Conflict]:
        """
        Checks if the setting can be changed without conflicts.
        Returns a string with the conflicts if there are any, otherwise returns
        an empty string.
        """

        setting_type, setting_attrs, get_settings, _ =\
            self.resolve_settings_type(setting_path_name)

        return Conflicts.detect(setting_type, setting_attrs, new_value,
                                get_settings())

    def get_setting_attr(self, setting_path_name: str) -> object:
        """Helper method to get the settings.

        In this case the setting_path_name can be a multi-layered setting, for example
        the kill switch can be access via `settings.killswtich` while netshield can be accessed
        via `settings.features.netshield`. Since this method tries to abstract the depth,
        we'll be searching based on the hierarchy, example:

        ```
        setting_path_name = "settings.features.netshield"
        setting_type = "settings"

        # The below will become ["features", "netshield"]
        setting_attrs_split = "features.netshield".split(".")
        ```
        So the for loop will loop for each attribute and attempt to get it from the original
        settings object, thus solving the nesting situation.

        Args:
            setting_path_name (str):

        Returns:
            object:
        """
        _, setting_attrs, get_settings, _ =\
            self.resolve_settings_type(setting_path_name)

        settings = get_settings()
        for attr in setting_attrs.split(DOT):
            settings = getattr(settings, attr)

        return settings

    def save_setting_attr(self, setting_path_name: str, new_value: object):
        """Helper method to save the settings."""
        def set_setting(root, attr, value):
            if attr.count(DOT) == 0:
                setattr(root, attr, value)
            else:
                name, path = attr.split(DOT, maxsplit=1)
                next_root = getattr(root, name)
                set_setting(next_root, path, value)

        setting_type, setting_attrs, get_settings, save_settings =\
            self.resolve_settings_type(setting_path_name)

        settings = get_settings()

        set_setting(settings, setting_attrs, new_value)

        save_settings(Conflicts.resolve(setting_type, setting_attrs, new_value,
                                        settings))

    def get_available_protocols(self, protocol_group: str) -> List[type[VPNConnection]]:
        """Returns a list of available protocols sorted by priority."""

        return list(self._connector.iter_available_protocols(protocol_group))

    def send_error_to_proton(
        self,
        error: Union[
            BaseException,
            tuple[
                Optional[Type[BaseException]],
                Optional[BaseException],
                Optional[TracebackType]]
            ]
    ):
        """Sends the error to Sentry."""
        self._api.usage_reporting.report_error(error)

    def run_subprocess(self, commands: list, check: bool = False) -> Future:
        """Run asynchronously subprocess command so it does not block UI."""
        return self.executor.submit(
            subprocess.run,
            commands,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=check,
        )

    @property
    def _local_agent_available(self) -> bool:
        try:
            import proton.vpn.local_agent  # noqa: F401,E501 # pylint: disable=unused-import, import-outside-toplevel
        except ModuleNotFoundError:
            return False

        return True

    def enable_refresher(self, callback: Callable[[Future], None]):
        """Enables the refresher and calls the callback when the server list is available."""
        def error_callback(exception: Exception):
            GLib.idle_add(
                self.exception_handler.handle_exception,
                type(exception), exception, exception.__traceback__
            )

        async def enable():
            self._api.refresher.set_error_callback(error_callback)
            await self._api.refresher.enable()

        future = self.executor.submit(enable)

        def on_refresher_enabled(future):
            GLib.idle_add(callback, future)

        future.add_done_callback(on_refresher_enabled)

    def disable_refresher(self):
        """Disables the refresher."""
        async def disable():
            await self._api.refresher.disable()
            self._api.refresher.unset_error_callback()

        future = self.executor.submit(disable)
        future.add_done_callback(lambda f: GLib.idle_add(f.result))

    def set_server_list_updated_callback(self, callback: Callable[[], None]):
        """Sets the callback that is called when the server list is updated."""
        future = self.executor.submit(
            self._api.refresher.set_server_list_updated_callback,
            lambda: GLib.idle_add(callback)
        )
        future.add_done_callback(lambda f: GLib.idle_add(f.result))

    def unset_server_list_updated_callback(self):
        """Unsets the callback that is called when the server list is updated."""
        future = self.executor.submit(self._api.refresher.set_server_list_updated_callback, None)
        future.add_done_callback(lambda f: GLib.idle_add(f.result))

    def set_server_loads_updated_callback(self, callback: Callable[[], None]):
        """Sets the callback that is called when the server loads are updated."""
        future = self.executor.submit(
            self._api.refresher.set_server_loads_updated_callback,
            lambda: GLib.idle_add(callback)
        )
        future.add_done_callback(lambda f: GLib.idle_add(f.result))

    def unset_server_loads_updated_callback(self):
        """Unsets the callback that is called when the server loads are updated."""
        future = self.executor.submit(self._api.refresher.set_server_loads_updated_callback, None)
        future.add_done_callback(lambda f: GLib.idle_add(f.result))

    @property
    def split_tunneling_available(self) -> bool:
        """Returns if split tunneling is available.

        Returns:
            bool: `True` if available, `False` otherwise
        """
        return self._connector.is_split_tunneling_available

    @property
    def fido2_available(self) -> bool:
        """
        Returns if FIDO2 is available.
        """
        return self._api.supports_fido2



"""
Connection settings module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import TYPE_CHECKING

from gi.repository import Gtk

from proton.vpn.app.gtk.conflicts import WIREGUARD_PROTOCOL

from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import (
    BaseCategoryContainer, BetaTag, ToggleWidget, ConflictableComboboxWidget,
    ReactiveSettingContainer, ReactiveSetting
)
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.custom_dns import CustomDNSWidget

if TYPE_CHECKING:
    from proton.vpn.app.gtk.widgets.headerbar.menu.settings.settings_window import \
        SettingsWindow


class ProtocolComboboxWidget(ConflictableComboboxWidget, ReactiveSetting):
    """Combobox widget for selecting the VPN protocol."""

    PROTUN_CHECKBOX_LABEL = "Use Proton protocols"

    PROTUN_PROTOCOL_GROUP = "protun"
    GENERIC_PROTOCOL_GROUP = "generic"

    def on_settings_changed(self, settings):
        """
        When settings are changed, we need to check if the protocol is
        still valid and update the combobox accordingly.
        """
        if self.combobox.get_active_text() != settings.protocol:
            self.combobox.set_active_id(settings.protocol)

    def get_protun_protocols(self) -> list[str]:
        """
        Returns the list of protun protocols if the feature flag is enabled,
        otherwise returns an empty list. We can use this method to determine
        whether the protun protocols should be shown in the UI or not.
        """
        # is_protun_enabled = self._controller.feature_flags.get("ProTunV1")
        is_protun_enabled = True
        if is_protun_enabled:
            return self._controller.get_available_protocols(self.PROTUN_PROTOCOL_GROUP)
        return []

    def protocol_group(self, current_protocol: str, protun_protocols: list) -> str:
        """Returns the protocol group for the given protocol."""
        for protocol in protun_protocols:
            if str(protocol.protocol) == current_protocol:
                return self.PROTUN_PROTOCOL_GROUP
        return self.GENERIC_PROTOCOL_GROUP

    def _build_ui(self):
        super()._build_ui()

        current_protocol = self.get_setting()
        protun_protocols = self.get_protun_protocols()

        if protun_protocols:
            current_protocol_group = self.protocol_group(current_protocol, protun_protocols)
            self._build_protun_checkbox(current_protocol_group)
        else:
            current_protocol_group = self.GENERIC_PROTOCOL_GROUP

        current_protocols = self._controller.get_available_protocols(current_protocol_group)
        self._repopulate_combobox(current_protocols)

        with self.pause_callback():
            self.combobox.set_active_id(current_protocol)

    def _build_protun_checkbox(self, current_protocol_group):
        """Builds and attaches the protun checkbox, initialising it from the current setting."""
        is_protun = current_protocol_group == self.PROTUN_PROTOCOL_GROUP

        protun_checkbox = Gtk.CheckButton(label=self.PROTUN_CHECKBOX_LABEL)
        protun_checkbox.set_active(is_protun)
        safe_signal_connect(protun_checkbox, "toggled", self._on_protun_checkbox_toggled)

        checkbox_row = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=8)
        checkbox_row.append(protun_checkbox)
        checkbox_row.append(BetaTag())
        self.attach(checkbox_row, 0, 2, 2, 1)

    def _repopulate_combobox(self, protocols):
        """Repopulates the combobox with the given protocols."""
        with self.pause_callback():
            self.combobox.set_entry_text_column(0)
            self.combobox.remove_all()
            for protocol in protocols:
                self.combobox.append(str(protocol.protocol), protocol.ui_protocol)
            self.combobox.set_entry_text_column(1)

    def _on_protun_checkbox_toggled(self, checkbox):
        if checkbox.get_active():
            protocol_group = self.PROTUN_PROTOCOL_GROUP
        else:
            protocol_group = self.GENERIC_PROTOCOL_GROUP
        protocols = self._controller.get_available_protocols(protocol_group)
        self._repopulate_combobox(protocols)
        if protocols:
            self.combobox.set_active(0)


class ConnectionSettings(BaseCategoryContainer, ReactiveSettingContainer):  # noqa: E501 # pylint: disable=line-too-long, too-many-instance-attributes
    """Settings related to connection are all grouped under this class."""
    CATEGORY_NAME = "Connection"
    PROTOCOL_LABEL = "Protocol"
    PROTOCOL_DESCRIPTION = "Protocol can only be changed when VPN is disconnected."
    VPN_ACCELERATOR_LABEL = "VPN Accelerator"
    VPN_ACCELERATOR_DESCRIPTION = "Increase your connection speed by up to 400% "\
        "with performance enhancing technologies."
    MODERATE_NAT_LABEL = "Moderate NAT"
    MODERATE_NAT_DESCRIPTION = "Disables randomization of the local addresses mapping. "\
        "This can slightly reduce connection security, but should allow direct "\
        "connections for online gaming and similar purposes."
    SWITCH_PROTOCOL_IF_CONNECTION_ACTIVE_DESCRIPTION = "Protocol selection "\
        "is disabled while VPN is active. Disconnect to make changes."
    IPV6_LABEL = "IPv6"
    IPV6_DESCRIPTION = "Tunnels IPv6 traffic through the VPN. "\
        "Can enhance compatibility with IPv6 networks."

    def __init__(self, controller: Controller, settings_window: "SettingsWindow"):
        super().__init__(self.CATEGORY_NAME)
        self._controller = controller
        self._settings_window = settings_window
        self.custom_dns = None
        self._protocol_widget = None
        self._vpn_accelerator_toggle = None
        self._moderate_nat_toggle = None
        self._ipv6_toggle = None

    def build_ui(self):
        """Builds the UI, invoking all necessary methods that are
        under this category."""
        self.build_protocol()
        self.build_vpn_accelerator()
        self.build_moderate_nat()
        self.build_ipv6()
        self.build_custom_dns()

    def build_protocol(self):
        """Builds and adds the `protocol` setting to the widget."""
        def do_set(combobox, new_value: str):
            combobox.save_setting(new_value)

        def do_revert(combobox):
            combobox.combobox.set_active_id(WIREGUARD_PROTOCOL)

        self._protocol_widget = ProtocolComboboxWidget(
            controller=self._controller,
            title=self.PROTOCOL_LABEL,
            description=self.PROTOCOL_DESCRIPTION,
            setting_name="settings.protocol",
            combobox_options=[],
            disable_on_active_connection=True,
            do_set=do_set,
            do_revert=do_revert
        )
        self.append(self._protocol_widget)

    def build_vpn_accelerator(self):
        """Builds and adds the `vpn_accelerator` setting to the widget."""
        def on_switch_state(_, new_value: bool, toggle_widget: ToggleWidget):
            toggle_widget.save_setting(new_value)

        self._vpn_accelerator_toggle = ToggleWidget(
            controller=self._controller,
            title=self.VPN_ACCELERATOR_LABEL,
            description=self.VPN_ACCELERATOR_DESCRIPTION,
            setting_name="settings.features.vpn_accelerator",
            requires_subscription_to_be_active=True,
            callback=on_switch_state
        )
        self.append(self._vpn_accelerator_toggle)

    def build_moderate_nat(self):
        """Builds and adds the `moderate_nat` setting to the widget."""
        def on_switch_state(_, new_value: bool, toggle_widget: ToggleWidget):
            toggle_widget.save_setting(new_value)

        self._moderate_nat_toggle = ToggleWidget(
            controller=self._controller,
            title=self.MODERATE_NAT_LABEL,
            description=self.MODERATE_NAT_DESCRIPTION,
            setting_name="settings.features.moderate_nat",
            requires_subscription_to_be_active=True,
            callback=on_switch_state
        )
        self.append(self._moderate_nat_toggle)

    def build_ipv6(self):
        """Builds and adds the `ipv6` setting to the widget."""
        def on_switch_state(_, new_value: bool, toggle_widget: ToggleWidget):
            toggle_widget.save_setting(new_value)
            self._settings_window.notify_user_with_reconnect_message(force_notify=True)

        self._ipv6_toggle = ToggleWidget(
            controller=self._controller,
            title=self.IPV6_LABEL,
            description=self.IPV6_DESCRIPTION,
            setting_name="settings.ipv6",
            callback=on_switch_state
        )
        self.append(self._ipv6_toggle)

    def build_custom_dns(self):
        """Builds and adds the `custom_dns` setting to the widget."""
        self.custom_dns = CustomDNSWidget.build(self._controller, self._settings_window)
        self.append(self.custom_dns)



"""
This module defines the main widget. The main widget is the widget which
exposes all the available app functionality to the user.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Union, TYPE_CHECKING, Optional

from proton.vpn.connection import states
from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.widgets.login.login_widget import LoginWidget
from proton.vpn.app.gtk.widgets.main.notification_bar import NotificationBar
from proton.vpn.app.gtk.widgets.vpn import VPNWidget
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget, DefaultLoadingWidget
from proton.vpn.app.gtk.widgets.main.notifications import Notifications
from proton.vpn.app.gtk.util import connect_once

if TYPE_CHECKING:
    from proton.vpn.app.gtk.controller import Controller
    from proton.vpn.app.gtk.app import MainWindow


# pylint: disable=too-many-instance-attributes
class MainWidget(Gtk.Overlay):
    """
    Main Proton VPN widget. It switches between the LoginWidget and the
    VPNWidget, depending on whether the user is logged in or not.
    """
    ERROR_DIALOG_PRIMARY_TEXT = "Something went wrong"
    SESSION_EXPIRED_ERROR_MESSAGE = "Your session has expired. "\
        "Please sign in again."
    SESSION_EXPIRED_ERROR_TITLE = "Invalid Session"

    def __init__(
        self, controller: "Controller", main_window: "MainWindow",
        overlay_widget: OverlayWidget, notifications: Optional[Notifications] = None
    ):
        super().__init__()
        self.set_name("main-widget")

        self.main_layout = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.main_layout.set_name("main-layout")

        self.content_layout = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.content_layout.set_name("content-layout")

        self._overlay_widget = overlay_widget
        self.main_layout.append(self.content_layout)
        self.set_child(self.main_layout)
        self.add_overlay(self._overlay_widget)

        self._active_widget = None
        self._controller = controller
        self._main_window = main_window
        self._connected_signals: list[tuple[int, Gtk.Widget]] = []

        self._notifications = notifications or Notifications(
            main_window, NotificationBar()
        )
        self.main_layout.prepend(self.notifications.notification_bar)
        self.login_widget = self._create_login_widget()
        self.vpn_widget = None

        safe_signal_connect(self, "realize", self._register_to_exception_handler)
        safe_signal_connect(self, "realize", self._init_on_realize)
        safe_signal_connect(self, "unrealize", self._unregister_from_exception_handler)
        safe_signal_connect(
            self._main_window.header_bar.menu,
            "user-logged-out",
            self._on_user_logged_out
        )

    def _register_to_exception_handler(self, *_):
        self._controller.exception_handler.main_widget = self

    def _unregister_from_exception_handler(self, *_):
        self._controller.exception_handler.main_widget = None

    def _init_on_realize(self, *_):
        self.initialize_visible_widget()

    @property
    def notifications(self) -> Notifications:
        """Returns the notifications object."""
        return self._notifications

    @property
    def active_widget(self):
        """Returns the active widget."""
        return self._active_widget

    @active_widget.setter
    def active_widget(self, widget: Union[LoginWidget, VPNWidget]):
        """Sets the active widget. That is, the widget to be shown
        to the user."""
        if self._active_widget:
            self.content_layout.remove(self._active_widget)
        self._active_widget = widget
        self.content_layout.append(self._active_widget)

    _STATE_CSS_CLASSES = {
        states.Connected: "vpn-connected",
        states.Connecting: "vpn-connecting",
        states.Disconnecting: "vpn-disconnecting",
        states.Disconnected: "vpn-disconnected",
        states.Error: "vpn-error",
    }

    def set_background_gradient(self, state: Optional[states.State]):
        """Sets the gradient on the main widget background for the given VPN state.

        :param state: a connection state instance, or None to clear all.
        """
        for css_class in self._STATE_CSS_CLASSES.values():
            self.remove_css_class(css_class)
        if state and (css_class := self._STATE_CSS_CLASSES.get(type(state))):
            self.add_css_class(css_class)

    def initialize_visible_widget(self):
        """
        Initializes the widget by showing either the vpn widget or the
        login widget depending on whether the user is authenticated or not.
        """
        if self._controller.user_logged_in:
            self._display_vpn_widget()
            connect_once(
                self.vpn_widget,
                "vpn-widget-ready",
                self._controller.run_startup_actions
            )
        else:
            self._display_login_widget()

    def show_error_message(
        self, error_message: str, blocking: bool = False,
        error_title: Optional[str] = None
    ):
        """
        Shows an error message to the user. The message is hidden after the
        specified amount of time.
        :param error_message: error message to be shown.
        :param blocking: whether the error message should require
        confirmation from the user or not.
        """
        if blocking:
            self.notifications.show_error_dialog(error_message, error_title)
        else:
            self.notifications.show_error_message(error_message)

    def on_session_expired(self):
        """This method is called by the exception handler once the session
        expires."""
        self.notifications.show_error_dialog(
            title=self.SESSION_EXPIRED_ERROR_TITLE,
            message=self.SESSION_EXPIRED_ERROR_MESSAGE
        )
        self._display_login_widget()

    def logout(self):
        """Logs out the user."""
        self._main_window.header_bar.menu.logout_button_click()

    def _on_user_logged_in(self, _login_widget: LoginWidget):
        self._display_vpn_widget()

    def _on_user_logged_out(self, *_):
        self._display_login_widget()

    def _hide_overlay_widget(self, *_):
        self._overlay_widget.hide()

    def _create_login_widget(self) -> LoginWidget:
        login_widget = LoginWidget(
            self._controller, self.notifications,
            self._overlay_widget, self._main_window
        )
        safe_signal_connect(login_widget, "user-logged-in", self._on_user_logged_in)
        return login_widget

    def _create_vpn_widget(self) -> VPNWidget:
        vpn_widget = VPNWidget(
            controller=self._controller,
            main_window=self._main_window,
            notifications=self.notifications
        )
        safe_signal_connect(
            vpn_widget, "vpn-widget-ready", self._hide_overlay_widget
        )
        signal_id = safe_signal_connect(
            vpn_widget,
            "connection-state-changed",
            self._change_gradient_on_connection_state_change
        )
        self._connected_signals.append((signal_id, vpn_widget))

        return vpn_widget

    def _change_gradient_on_connection_state_change(self, _, state):
        self.set_background_gradient(state)

    def _display_vpn_widget(self):
        self.vpn_widget = self._create_vpn_widget()
        self._main_window.header_bar.menu.logout_enabled = True
        self._main_window.header_bar.menu.settings_enabled = True
        self._overlay_widget.show(
            DefaultLoadingWidget("Loading app...")
        )
        self.active_widget = self.vpn_widget
        self.vpn_widget.load()

    def _display_login_widget(self):
        self.set_background_gradient(None)
        for signal_id, widget in self._connected_signals:
            widget.disconnect(signal_id)
        self._connected_signals.clear()
        self._main_window.header_bar.menu.logout_enabled = False
        self._main_window.header_bar.menu.settings_enabled = False
        # Close the settings window in case the session expires with the settings windows open.
        self._main_window.header_bar.menu.close_settings_window()
        self._overlay_widget.hide()  # Required on session expired while loading VPN widget.
        self.active_widget = self.login_widget
        self.login_widget.reset()



"""
This module defines the widgets required to display the form to
authenticate with username and password.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from concurrent.futures import Future

from gi.repository import GLib, GObject

from proton.vpn import logging

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.login.logo import ProtonVPNLogo
from proton.vpn.app.gtk.widgets.login.password_entry import PasswordEntry
from proton.vpn.app.gtk.widgets.main.notifications import Notifications
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget, DefaultLoadingWidget

logger = logging.getLogger(__name__)


class LoginForm(Gtk.Box):  # pylint: disable=R0902
    """It implements the login form. Once the user is authenticated, it
    emits the `user-authenticated` signal.

    Note that 2FA is not implemented by this widget. For that see
    TwoFactorAuthForm.

    """
    LOGGING_IN_MESSAGE = "Signing in..."
    INVALID_USERNAME_MESSAGE = "Invalid username."
    INCORRECT_CREDENTIALS_MESSAGE = "Incorrect credentials."

    def __init__(
        self,
        controller: Controller,
        notifications: Notifications,
        overlay_widget: OverlayWidget
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=30)
        self.set_name("login-form")
        self.set_vexpand(True)
        self.set_halign(Gtk.Align.FILL)
        self.set_valign(Gtk.Align.FILL)
        self._controller = controller
        self._notifications = notifications
        self._overlay_widget = overlay_widget

        self.append(ProtonVPNLogo())

        self._username_entry = Gtk.Entry()
        self._username_entry.set_placeholder_text("Username")
        self._username_entry.set_input_purpose(Gtk.InputPurpose.FREE_FORM)
        self.append(self._username_entry)

        self._password_entry = PasswordEntry()
        self._password_entry.set_placeholder_text("Password")
        self.append(self._password_entry)

        self._login_button = Gtk.Button(label="Sign in")
        safe_signal_connect(self._login_button, "clicked", self._on_login_button_clicked)
        self._login_button.add_css_class("primary")
        self._login_button.add_css_class("spaced")
        self._login_button.set_halign(Gtk.Align.CENTER)
        # By default, the button should never be clickable, as username and
        # password fields are empty and users need to actively provide an input
        # to unlock the login button.
        self._login_button.set_property("sensitive", False)
        self.append(self._login_button)

        # Listen to key entries so that the login button can be "unlocked"
        # once username and password are provided.
        safe_signal_connect(
            self._password_entry, "changed", self._on_entry_changed
        )
        safe_signal_connect(
            self._username_entry, "changed", self._on_entry_changed
        )

        # Allows both entries to react to 'Enter' button
        safe_signal_connect(self._username_entry, "activate", self._on_press_enter)
        safe_signal_connect(self._password_entry, "activate", self._on_press_enter)

        self.append(LoginLinks())

    def _on_press_enter(self, _):
        if not self._login_button.get_property("sensitive"):
            return

        self._login_button.emit("clicked")

    def _on_login_button_clicked(self, _):
        logger.info("Clicked on login", category="UI", subcategory="LOGIN", event="CLICK")
        self._overlay_widget.show(
            DefaultLoadingWidget(self.LOGGING_IN_MESSAGE)
        )
        future = self._controller.login(self.username, self.password)
        future.add_done_callback(
            lambda future: GLib.idle_add(self._on_login_result, future)
        )

    def _on_login_result(self, future: Future):
        try:
            result = future.result()
        except ValueError as error:
            self._notifications.show_error_message(self.INVALID_USERNAME_MESSAGE)
            logger.warning(
                error, category="APP", subcategory="LOGIN", event="RESULT",
                exc_info=True
            )
            self.emit("login-error")
            return
        finally:
            self._overlay_widget.hide()

        if result.authenticated:
            self._signal_user_authenticated(result.twofa_required)
        else:
            self._notifications.show_error_message(self.INCORRECT_CREDENTIALS_MESSAGE)
            logger.warning(
                self.INCORRECT_CREDENTIALS_MESSAGE, category="APP",
                subcategory="LOGIN", event="RESULT"
            )
            self.emit("login-error")

    def _on_entry_changed(self, _):
        """Toggles login button state based on username and password lengths."""
        is_username_provided = len(self.username.strip()) > 0
        is_password_provided = len(self.password.strip()) > 0
        is_data_provided = is_username_provided and is_password_provided

        self._login_button.set_property("sensitive", is_data_provided)

    def _signal_user_authenticated(self, two_factor_auth_required: bool):
        self.emit("user-authenticated", two_factor_auth_required)

    @GObject.Signal(name="user-authenticated", arg_types=(bool,))
    def user_authenticated(self, two_factor_auth_required: bool):
        """
        Signal emitted after the user successfully authenticates.
        :param two_factor_auth_required: whether 2FA is required or not.
        """

    @GObject.Signal(name="login-error")
    def login_error(self):
        """Signal emitted when a login error occurred."""

    def reset(self):
        """Resets the state of the login/2fa forms."""
        self._notifications.hide_message()
        self.username = ""
        self.password = ""  # nosec B105
        self._username_entry.grab_focus()

    @property
    def error_message(self) -> str:
        """Return the contents of the error message in the notification bar."""
        return self._notifications.notification_bar.current_message

    @property
    def username(self) -> str:
        """Returns the username introduced in the login form."""
        return self._username_entry.get_text()

    @username.setter
    def username(self, username: str):
        """Sets the username in the login form."""
        self._username_entry.set_text(username)

    @property
    def password(self) -> str:
        """Returns the password introduced in the login form."""
        return self._password_entry.get_text()

    @password.setter
    def password(self, password: str):
        """Sets the password in the login form."""
        self._password_entry.set_text(password)

    @property
    def is_login_button_clickable(self) -> bool:
        """Check if the login button is clickable or not.
        This property was made available mainly for testing purposes."""
        return self._login_button.get_property("sensitive")

    def submit_login(self):
        """Submits the login form.
        This property was made available mainly for testing purposes."""
        self._login_button.emit("clicked")

    def username_enter(self):
        """Submits the login form from the username entry.
        This property was made available mainly for testing purposes."""
        self._username_entry.emit("activate")

    def password_enter(self):
        """Submits the login form from the password entry.
        This property was made available mainly for testing purposes."""
        self._password_entry.emit("activate")


class LoginLinks(Gtk.Box):
    """Links shown in the login widget."""
    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        self.set_vexpand(True)
        self.set_valign(Gtk.Align.END)
        create_account_link = Gtk.LinkButton(
            label="Create Account",
            uri="https://account.protonvpn.com/signup?ref=linux"
        )
        create_account_link.set_halign(Gtk.Align.START)
        create_account_link.set_hexpand(True)
        self.append(create_account_link)
        help_link = Gtk.LinkButton(
            label="Need Help?",
            uri="https://protonvpn.com/support"
        )
        help_link.set_halign(Gtk.Align.END)
        help_link.set_hexpand(True)
        self.append(help_link)



"""
This module defines the widget used to display the security device form.


Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import threading
from concurrent.futures import Future
from typing import Optional

from gi.repository import GLib, GObject

from proton.vpn import logging
from proton.vpn.session.exceptions import \
    SecurityKeyError, SecurityKeyNotFoundError, InvalidSecurityKeyError, \
    SecurityKeyPINNotSetError, SecurityKeyPINInvalidError, Fido2NotSupportedError

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.login.logo import SecurityKeyLogo
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import SettingDescription
from proton.vpn.app.gtk.widgets.login.password_entry import PasswordEntry
from proton.vpn.app.gtk.widgets.login.two_factor_auth.authenticate_button import AuthenticateButton
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget
from proton.vpn.app.gtk.widgets.main.notifications import Notifications

logger = logging.getLogger(__name__)

LEARN_MORE_LINK = '<a href="https://protonvpn.com/support/#">Learn more</a>'


class SecurityKeyForm(Gtk.Box):  # pylint: disable=R0902
    """
    Implements the UI for HOTP authentication.
    Once the HOTP code is authenticated,
    it emits the `hotp-auth-successful` signal.
    """
    DESCRIPTION_LABEL = "Insert the U2F or FIDO key linked to your Proton Account. " \
        + LEARN_MORE_LINK
    PIN_CODE_LABEL = "PIN code"

    MULTIPLE_SECURITY_KEYS_FOUND = "Multiple security keys were found. Tap one to select it."
    PHYSICAL_VERIFICATION_MESSAGE = "If your security key has a button or a gold disc, tap it now."
    SECURITY_KEY_NOT_FOUND_MESSAGE = "Two-factor authentication failed: No security key detected"
    INVALID_SECURITY_KEY_MESSAGE = "Two-factor authentication failed: The security key you used " \
        "is not linked to your Proton Account."
    FIDO2_NOT_SUPPORTED_MESSAGE = "Two-factor authentication failed. Security key 2FA is not " \
        "enabled for your account."
    SECURITY_KEY_PIN_NOT_SET_MESSAGE = "Two-factor authentication failed: " \
        "Your security key has no PIN set"
    SECURITY_KEY_PIN_INVALID_MESSAGE = "Two-factor authentication failed: Incorrect PIN"
    GENERIC_ERROR_MESSAGE = "An unknown error occurred"
    LOGGING_IN_MESSAGE = "Signing in..."

    def __init__(
            self,
            controller: Controller,
            notifications: Notifications,
            overlay_widget: OverlayWidget,
            authenticate_button: Optional[AuthenticateButton] = None
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=15)

        self.set_name("security-device-form")

        self._controller = controller
        self._notifications = notifications
        self._overlay_widget = overlay_widget

        self._authenticate_button = authenticate_button or AuthenticateButton()
        safe_signal_connect(
            self._authenticate_button, "clicked", self._on_authenticate_button_clicked
        )

        self._cancel_button = Gtk.Button(label="Cancel")
        self._cancel_button.add_css_class("danger")
        safe_signal_connect(self._cancel_button, "clicked", self._on_cancel_button_clicked)

        self._instruction_label = SettingDescription(self.DESCRIPTION_LABEL)
        self._instruction_label.remove_css_class("dim-label")
        self._instruction_label.set_wrap(True)

        self._pin_code_label = Gtk.Label(label=self.PIN_CODE_LABEL)
        self._pin_code_label.set_halign(Gtk.Align.START)

        self._pin_code_entry = PasswordEntry()
        self._pin_code_entry.set_input_purpose(Gtk.InputPurpose.FREE_FORM)
        safe_signal_connect(
            self._pin_code_entry, "changed", self._on_pin_code_entry_changed
        )
        safe_signal_connect(
            self._pin_code_entry, "activate", self._on_authenticate_button_clicked
        )

        self._requesting_pin: Optional[threading.Event] = None
        self._cancel_assertion: Optional[threading.Event] = None

        # Box is used to group the pin code label and the pin code entry.
        self._pin_code_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, spacing=10
        )
        self._pin_code_box.append(self._pin_code_label)
        self._pin_code_box.append(self._pin_code_entry)

        self._pin_code_box_revealer = Gtk.Revealer()
        self._pin_code_box_revealer.set_name("pin-code-box-revealer")
        self._pin_code_box_revealer.set_child(self._pin_code_box)
        safe_signal_connect(
            self._pin_code_box_revealer,
            "notify::reveal-child", self._on_pin_code_box_revealer_notify_reveal_child
        )

        # Box is used to group the pin code box revealer and the authenticate button.
        self._authenticate_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=10)
        self._authenticate_box.set_vexpand(True)
        self._authenticate_box.set_valign(Gtk.Align.END)
        self._authenticate_box.append(self._pin_code_box_revealer)
        self._authenticate_box.append(self._authenticate_button)
        self._authenticate_box.append(self._cancel_button)
        self.append(self._instruction_label)
        self.append(SecurityKeyLogo())
        self.append(self._authenticate_box)

    @GObject.Signal
    def two_factor_auth_cancelled(self):
        """Signal emitted after the user cancelled 2FA"""

    @GObject.Signal
    def two_factor_auth_successful(self):
        """Signal emitted after a successful 2FA."""

    def _on_authenticate_button_clicked(self, _):
        """Called when the authenticate button is clicked."""
        self._authenticate_button.set_sensitive(False)
        if self._requesting_pin:
            # Authenticate button clicked again after PIN request
            self._requesting_pin.set()
        else:
            self._cancel_assertion = threading.Event()
            logger.info(
                "Clicked on authenticate via security key",
                category="UI", subcategory="LOGIN-2FA", event="CLICK"
            )
            future = self._controller.generate_2fa_fido2_assertion(
                user_interaction=self,
                cancel_assertion=self._cancel_assertion
            )
            future.add_done_callback(
                lambda future: GLib.idle_add(self._on_2fa_fido2_assertion,
                                             future)
            )

    def _on_cancel_button_clicked(self, _):
        if self._cancel_assertion:
            self._cancel_assertion.set()
        self.emit("two-factor-auth-cancelled")
        self.reset()

    def _on_2fa_fido2_assertion(self, future: Future):
        msg = None

        try:
            assertion = future.result()
        except Fido2NotSupportedError:
            msg = self.FIDO2_NOT_SUPPORTED_MESSAGE
        except SecurityKeyNotFoundError:
            msg = self.SECURITY_KEY_NOT_FOUND_MESSAGE
        except InvalidSecurityKeyError:
            msg = self.INVALID_SECURITY_KEY_MESSAGE
        except SecurityKeyPINNotSetError:
            msg = self.SECURITY_KEY_PIN_NOT_SET_MESSAGE
        except SecurityKeyPINInvalidError:
            msg = self.SECURITY_KEY_PIN_INVALID_MESSAGE
        except (SecurityKeyError) as excp:
            logger.exception(str(excp), category="APP", subcategory="LOGIN-2FA", event="ERROR")
            msg = self.GENERIC_ERROR_MESSAGE

        self.reset()

        if msg:
            self._notifications.show_error_message(msg)
            logger.warning(msg, category="APP", subcategory="LOGIN-2FA", event="ERROR")
            self._overlay_widget.hide()
            return

        self._overlay_widget.show_message(self.LOGGING_IN_MESSAGE)

        future = self._controller.submit_2fa_fido2(assertion)
        future.add_done_callback(
            lambda future: GLib.idle_add(self._on_2fa_fido2_submission_result, future)
        )

    def _on_2fa_fido2_submission_result(self, future: Future):
        try:
            result = future.result()
        finally:
            self._overlay_widget.hide()

        if result.success:
            self.emit("two-factor-auth-successful")
        else:
            self._notifications.show_error_message(self.GENERIC_ERROR_MESSAGE)

    def _on_pin_code_box_revealer_notify_reveal_child(self, *_):
        """Called when the pin code box revealer is notified of a reveal child change."""
        self._authenticate_button.enable = len(self._pin_code_entry.get_text()) > 0

    def _on_pin_code_entry_changed(self, _):
        """Toggles pin code entry state based on pin code length."""
        self._authenticate_button.enable = len(self._pin_code_entry.get_text()) > 0

    def reset(self):
        """Resets the widget to its initial state."""
        self._pin_code_box_revealer.set_reveal_child(False)
        self._pin_code_entry.set_text("")
        self._authenticate_button.enable = True
        self._requesting_pin = None
        self._authenticate_button.set_sensitive(True)
        self._authenticate_button.grab_focus()

    def reveal_pin_code_entry(self):
        """Reveals the pin code entry."""
        self._pin_code_box_revealer.set_reveal_child(True)
        self._pin_code_entry.grab_focus()

    @property
    def authenticate_button_enabled(self):
        """Returns if the authenticate button is enabled."""
        return self._authenticate_button.enable

    def set_pin_code(self, pin_code: str):
        """Sets the pin code entry."""
        self._pin_code_entry.set_text(pin_code)

    def authenticate_button_click(self):
        """Clicks the authenticate button."""
        self._authenticate_button.emit("clicked")

    def request_key_selection(self):
        """Called when multiple keys are found and the user needs to select one
        by touching it."""
        GLib.idle_add(self._overlay_widget.show_message, self.MULTIPLE_SECURITY_KEYS_FOUND)

    def prompt_up(self) -> None:
        """Called when the FIDO2 client checks for user presence."""
        GLib.idle_add(self._overlay_widget.show_message, self.PHYSICAL_VERIFICATION_MESSAGE)

    def request_pin(self, *_args, **_kwargs) -> Optional[str]:
        """
        Called if the FIDO2 client requires a PIN.
        :returns: the PIN the user typed or None/empty to cancel.
        """
        self._requesting_pin = threading.Event()
        GLib.idle_add(self.reveal_pin_code_entry)
        GLib.idle_add(self._overlay_widget.hide)
        self._requesting_pin.wait()
        self._requesting_pin = None
        return self._pin_code_entry.get_text()

    def request_uv(self, *_args, **_kwargs) -> bool:
        """
        Called when the FIDO2 client is about to request UV (user verification) from the user.
        :returns: True if allowed, or False to cancel.
        """
        return True



"""
Auto reconnect feature.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import random
from typing import Optional

from gi.repository import GLib
from proton.vpn.core.refresher import VPNDataRefresher

from proton.vpn import logging
from proton.vpn.connection import states, VPNConnection, events
from proton.vpn.connection.exceptions import VPNConnectionError, \
    AuthenticationError, HardJailedTwoFAError, NotYetValidCertificateError
from proton.vpn.core.vpnconnector import VPNConnector

from proton.vpn.app.gtk.services.reconnector.network_monitor import NetworkMonitor
from proton.vpn.app.gtk.services.reconnector.session_monitor import SessionMonitor
from proton.vpn.app.gtk.services.reconnector.vpn_monitor import VPNMonitor
from proton.vpn.app.gtk.utils.executor import AsyncExecutor

logger = logging.getLogger(__name__)


class VPNReconnector:  # pylint: disable=too-many-instance-attributes
    """
    It implements the auto reconnect feature.

    The reconnector is in charge of reconnecting the current Proton VPN connection
    when it drops.

    Currently, it requires a GLib MainLoop to be running. In a future version,
    the reconnector will be refactored so that it runs in a separate python
    process and will run its own main loop.
    """

    # pylint: disable=too-many-arguments
    def __init__(
            self,
            vpn_connector: VPNConnector,
            vpn_data_refresher: VPNDataRefresher,
            vpn_monitor: VPNMonitor,
            network_monitor: NetworkMonitor,
            session_monitor: SessionMonitor,
            async_executor: AsyncExecutor
    ):
        self._vpn_connector = vpn_connector
        self._vpn_data_refresher = vpn_data_refresher

        self._vpn_monitor = vpn_monitor
        self._vpn_monitor.vpn_drop_callback = self._on_vpn_drop
        self._vpn_monitor.vpn_up_callback = self._on_vpn_up
        self._vpn_monitor.vpn_disconnected_callback = self._on_vpn_disconnected

        self._network_monitor = network_monitor
        self._network_monitor.network_up_callback = self._on_network_up

        self._session_monitor = session_monitor or SessionMonitor()
        self._session_monitor.session_unlocked_callback = self._on_session_unlocked

        self._executor = async_executor

        self._new_certificate_src_id: Optional[int] = None
        self._retry_src_id: Optional[int] = None
        self.retry_counter = 0

    @property
    def is_reconnection_scheduled(self) -> bool:
        """Returns True if there is a pending scheduled reconnection and False otherwise."""
        return self._retry_src_id is not None

    def enable(self):
        """Enables the auto reconnect feature."""
        if not self._vpn_data_refresher.is_vpn_data_ready:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            raise RuntimeError("VPN data refresher is not ready.")
        self._reset_retry_counter()
        self._vpn_monitor.enable()
        self._network_monitor.enable()
        self._session_monitor.enable()
        logger.info("VPN reconnector enabled.")

    def disable(self):
        """Disables the auto reconnect feature."""
        self._vpn_monitor.disable()
        self._network_monitor.disable()
        self._session_monitor.disable()
        logger.info("VPN reconnector disabled.")

    @property
    def did_vpn_drop(self) -> bool:
        """Returns True if the VPN connection dropped or False otherwise."""
        return isinstance(self._vpn_connector.current_state, states.Error)

    @property
    def is_connection_error_fatal(self) -> bool:
        """
        Returns True if a VPN reconnection is possible or False otherwise.
        """
        return (
            isinstance(self._vpn_connector.current_state, states.Error)
            and not isinstance(self._vpn_connector.current_state.context.event, events.AuthDenied)
        )

    def schedule_reconnection(self) -> bool:
        """Schedules a reconnection attempt.

        The amount of time elapsed before the reconnection is attempted
        depends on the number of previous failed reconnection attempts.

        :return: True if the reconnection could be scheduled and False otherwise.
        """
        if self._retry_src_id:
            logger.warning("There is already a scheduled VPN reconnection attempt.")
            return False

        retry_delay = self._calculate_retry_delay_in_milliseconds()
        logger.info(
            f"Reconnection attempt #{self.retry_counter} scheduled in "
            f"{retry_delay/1000:.2f} seconds.")
        self._retry_src_id = GLib.timeout_add(retry_delay, self._reconnect)
        return True

    @property
    def _current_connection(self) -> VPNConnection:
        return self._vpn_connector.current_connection

    def _on_reconnection_error(self):
        self._reset_retry_counter()

        event = self._vpn_connector.current_state.context.event
        if isinstance(event, events.AuthDenied):
            raise AuthenticationError("Reconnection not possible due to authentication error.")

        raise VPNConnectionError(f"Reconnection not possible due to unexpected event: {event}")

    def _on_two_fa_required(self):
        raise HardJailedTwoFAError(
            "Two factor authentication required to reconnect."
        )

    def _on_not_yet_valid_certificate(self):
        raise NotYetValidCertificateError(
            "The certificate is not valid yet."
        )

    def _on_session_unlocked(self):
        """
        Callback called by the session monitor once the user session has been
        unlocked.
        """
        logger.info("Session unlocked.")
        self._reset_retry_counter()

        if not self.did_vpn_drop:
            logger.debug("VPN reconnection not necessary: connection didn't drop.")
            return

        if not self.is_connection_error_fatal:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            logger.debug("VPN reconnection not possible: fatal connection error.")
            return

        self.schedule_reconnection()

    def _on_network_up(self):
        """
        Callback called by the network monitor once the machine's network state
        reaches the connected state, meaning that from now on is able to reach
        the internet.
        """
        logger.info("Network connectivity was detected.")
        self._reset_retry_counter()

        if not self.did_vpn_drop:
            logger.debug("VPN reconnection not necessary: connection didn't drop.")
            return

        if not self.is_connection_error_fatal:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            logger.debug("VPN reconnection not possible: fatal connection error.")
            return

        self.schedule_reconnection()

    def _on_vpn_drop(self, event: events.Event):
        """Callback called by the VPN monitor when a VPN connection drop was detected."""
        logger.info("VPN connection drop was detected.")
        if isinstance(event, events.ExpiredCertificate):
            self._handle_certificate_expired()
            return

        if isinstance(event, events.MaximumSessionsReached):
            return

        if isinstance(event, events.TwoFARequired):
            GLib.idle_add(self._on_two_fa_required)
            return

        if isinstance(event, events.NotYetValidCertificate):
            GLib.idle_add(self._on_not_yet_valid_certificate)
            return

        if not self.is_connection_error_fatal:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            logger.info("VPN reconnection not possible: fatal connection error.")
            # Raise exception on the next event loop iteration so that the app reacts to it.
            GLib.idle_add(self._on_reconnection_error)
            return

        self.schedule_reconnection()

    def _handle_certificate_expired(self):
        self._executor.submit(self._vpn_data_refresher.force_refresh_certificate)

    def _on_vpn_up(self):
        """Callback called by the VPN monitor when the VPN connection is up."""
        logger.debug("VPN connection is up.")
        self._reset_retry_counter()

    def _on_vpn_disconnected(self):
        """Callback called by the VPN monitor when the VPN connection is disconnected."""
        logger.info("VPN connection is disconnected.")
        self._reset_retry_counter()

    def _reconnect(self):
        logger.info(f"Reconnecting (attempt #{self.retry_counter})...")
        connection = self._vpn_connector.current_connection

        if not self._network_monitor.is_network_up:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            logger.info("VPN reconnection not possible: network is down.")
            self._increase_retry_counter()
            self.schedule_reconnection()
            return False

        if not self._session_monitor.is_session_unlocked:  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
            logger.info("VPN reconnection not possible: session is locked.")
            self._increase_retry_counter()
            self.schedule_reconnection()
            return False

        vpn_server = self._get_vpn_server(connection.server_id)
        if vpn_server:
            future = self._executor.submit(
                self._vpn_connector.connect,
                vpn_server,
                connection.protocol,
                connection.backend
            )
            future.add_done_callback(lambda f: GLib.idle_add(f.result))
            self._increase_retry_counter()
        else:
            # The server was removed from the server list after the user had connected to it.
            logger.warning(
                "VPN Reconnection not possible: logical server not found "
                f"(id = {connection.server_id})"
            )

        return False  # Remove periodic source

    def _get_vpn_server(self, server_id: str):
        logical_server = self._vpn_data_refresher.server_list.get_by_id(server_id)
        if not logical_server:
            return None
        client_config = self._vpn_data_refresher.client_config

        return self._vpn_connector.get_vpn_server(logical_server, client_config)

    def _calculate_retry_delay_in_milliseconds(self) -> int:
        """
        Returns the amount of milliseconds to wait before a VPN connection retry.

        The amount of time increases exponentially based on the number of
        previous attempts.
        """
        return (2 ** self.retry_counter *
                random.uniform(0.9, 1.1) * 1000)  # nosec B311 # noqa: E501 # pylint: disable=line-too-long # nosemgrep: gitlab.bandit.B311

    def _reset_retry_counter(self):
        if self._retry_src_id:
            GLib.source_remove(self._retry_src_id)
            self._retry_src_id = None
        self.retry_counter = 0

    def _increase_retry_counter(self):
        self.retry_counter += 1
        self._retry_src_id = None



"""
Different services running in the background.

Currently, these services are all running on the main (GLib) event loop.
However, the goal is to extract some (like the reconnector), to a systemd service
running on separate process. But, to be able to do that, first we need a VPN daemon
process coordinating the creation/deletion of VPN connections requested by other
processes like the app, the CLI or the reconnector, once extracted to a separate
process.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from proton.vpn.app.gtk.services.reconnector.reconnector import VPNReconnector

__all__ = ["VPNReconnector"]



"""
Toggleable "Via Secure Core" row for the country, listing secure core servers.

Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from typing import List, Optional

from gi.repository import GLib

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.session.dataclasses.servers import SecureCoreGroup
from proton.vpn.session.servers import LogicalServer, TierEnum

from proton.vpn.app.gtk.utils.assertions import runtime_assert
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.expandable_row import ExpandableRow
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.row_content import RowContent
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.row_view_model import RowViewModel
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.utils import (
    make_connect_callback, sync_rows_with_model_items
)
from proton.vpn.app.gtk.widgets.vpn.serverlist.icons import (
    DoubleFlagIcon,
    SecureCoreIcon,
)


class SecureCoreRow(Gtk.Box):
    """Toggleable row with label "Via Secure Core" that expands to show secure core servers."""
    LABEL = "Via Secure Core"

    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self._secure_core_group: Optional[SecureCoreGroup] = None
        self._controller: Optional[Controller] = None
        self._user_tier: Optional[int] = None
        self._expandable_row = ExpandableRow(
            on_expand=self._add_server_rows,
            on_collapse=self._remove_server_rows,
        )
        self.append(self._expandable_row)
        self._server_rows: List[RowContent] = []

    # pylint: disable=too-many-arguments
    def display(
        self,
        controller: Controller,
        secure_core_group: SecureCoreGroup,
        user_tier: int,
        expanded: bool = False,
    ) -> None:
        """Displays the secure core row according to the specified parameters."""
        self._controller = controller
        self._secure_core_group = secure_core_group
        self._user_tier = user_tier
        self._expandable_row.reset(keep_children=False)
        self._expandable_row.connect_toggle()
        exit_country_name = secure_core_group.servers[0].exit_country_name
        upgrade_required = user_tier == TierEnum.FREE and not secure_core_group.free
        connect_button_tooltip = (
            f"Upgrade to connect to {exit_country_name} via Secure Core"
            if upgrade_required else
            f"Connect to {exit_country_name} via Secure Core"
        )
        toggle_button_tooltips = (
            f"Show all Secure Core servers\nto connect to {exit_country_name}",
            f"Hide all Secure Core servers\nto connect to {exit_country_name}"
        )

        row_data = RowViewModel(
            name=self.LABEL,
            on_connect=make_connect_callback(controller, secure_core_group.servers, user_tier),
            free=secure_core_group.free,
            under_maintenance=secure_core_group.under_maintenance and not upgrade_required,
            features=set(),
            smart_routing=False,
            toggable=True,
            upgrade_required=upgrade_required,
            icon_factory=lambda: SecureCoreIcon(size=24),
            connect_button_tooltip=connect_button_tooltip,
            toggle_button_tooltips=toggle_button_tooltips,
        )
        self._expandable_row.row_content.display(row_data)
        if expanded:
            self._expandable_row.row_content.click_toggle_button()

    @property
    def row_content(self) -> RowContent:
        """Returns the header row content widget."""
        return self._expandable_row.row_content

    @property
    def server_rows(self) -> List[RowContent]:
        """Returns the list of server rows currently displayed."""
        return list(self._server_rows)

    @property
    def expanded(self) -> bool:
        """Returns whether the secure core row is currently expanded or not."""
        return self._expandable_row.row_content.expanded

    @property
    def label(self) -> str:
        """Returns the label of the secure core row."""
        return self._expandable_row.row_content.label

    def reset(self, keep_children: bool = False) -> None:
        """Resets the secure core row to its initial state."""
        self._expandable_row.reset(keep_children=keep_children)

    def _remove_server_rows(self) -> None:
        while self._server_rows:
            server_row = self._server_rows.pop()
            self._expandable_row.remove_child(server_row)
            server_row.reset()

    def _add_server_rows(self) -> None:
        runtime_assert(self._secure_core_group is not None, "Secure core group is not set")
        # Capture controller directly to avoid closing over `self` in on_connect
        controller = self._controller

        # pylint: disable=duplicate-code
        def display_server_row(server_row: RowContent, server: LogicalServer) -> None:
            upgrade_required = self._user_tier == TierEnum.FREE and not server.free

            def on_connect():
                future = controller.connect_to_server(server.name)
                future.add_done_callback(lambda f: GLib.idle_add(f.result))

            row_data = RowViewModel(
                name=f"Via {server.entry_country_name}",
                on_connect=on_connect,
                free=server.free,
                under_maintenance=server.under_maintenance and not upgrade_required,
                features=set(),
                smart_routing=False,
                toggable=False,
                upgrade_required=upgrade_required,
                load=None if server.under_maintenance else server.load,
                icon_factory=lambda s=server: DoubleFlagIcon(
                    exit_country_code=s.exit_country,
                    entry_country_code=s.entry_country,
                ),
                connect_button_tooltip=(
                    f"Upgrade to connect to {server.exit_country_name}"
                    f" via {server.entry_country_name}"
                    if upgrade_required else
                    f"Connect to {server.exit_country_name}"
                    f" via {server.entry_country_name}"
                ),
            )
            server_row.display(row_data)

        sync_rows_with_model_items(
            list(self._secure_core_group.servers),
            self._server_rows,
            self._expandable_row.container,
            RowContent,
            display_server_row,
        )



"""
This module defines the country rows displayed in the server list widget.


Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from itertools import chain
from typing import List, Optional

from proton.vpn.session.servers import Country, Location, TierEnum
from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.location_row import LocationRow
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.expandable_row import ExpandableRow
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.row_view_model import RowViewModel
from proton.vpn.app.gtk.utils.assertions import runtime_assert
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.utils \
    import make_connect_callback, sync_rows_with_model_items
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.secure_core_row import SecureCoreRow
from proton.vpn.app.gtk.widgets.vpn.serverlist.icons import CountryFlagIcon


# pylint: disable=too-many-instance-attributes
class CountryRow(Gtk.Box):
    """Row representing a country in the server list widget."""

    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self._country: Optional[Country] = None
        self._controller = None
        self._user_tier = None
        self._expanded_groups: set[str] = set()
        self._expandable_row = ExpandableRow(
            on_expand=self._on_expand,
            on_collapse=self._on_collapse,
        )
        self.append(self._expandable_row)
        self._location_row_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self._location_row_container.set_spacing(5)
        self._expandable_row.container.append(self._location_row_container)
        self._secure_core_row_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self._secure_core_row_container.set_spacing(5)
        self._expandable_row.container.append(self._secure_core_row_container)

        self._location_rows: List[LocationRow] = []
        self._secure_core_row: Optional[SecureCoreRow] = None

    # pylint: disable=too-many-arguments
    def display(
        self, controller: Controller, country: Country, user_tier: int,
        expanded: bool = False, expanded_groups: Optional[set[str]] = None
    ):
        """Displays the country row according to the specified parameters.

        Args:
            controller: The controller instance
            country: The country to display
            user_tier: The user's tier level
            expanded: Whether the country should be expanded (defaults to False)
            expanded_groups: Optional set of child group labels (lowercase)
                that should be expanded
        """
        expanded_groups = expanded_groups or set()
        self.reset(keep_location_rows=expanded)
        self._controller = controller
        self._country = country
        self._user_tier = user_tier
        self._expanded_groups = expanded_groups
        self._expandable_row.connect_toggle()
        upgrade_required = user_tier == TierEnum.FREE and not country.free

        row_data = RowViewModel(
            name=country.name,
            on_connect=make_connect_callback(controller, country.servers, user_tier),
            free=country.free,
            under_maintenance=country.under_maintenance and not upgrade_required,
            features=country.features,
            smart_routing=country.smart_routing,
            toggable=True,
            upgrade_required=upgrade_required,
            icon_factory=lambda c=country: CountryFlagIcon(c.code),
            connect_button_tooltip=(
                f"Upgrade to connect to {country.name}"
                if upgrade_required else
                f"Connect to {country.name}"
            ),
            toggle_button_tooltips=(
                f"Show all locations from {country.name}",
                f"Hide all locations from {country.name}",
            ),
        )
        self._expandable_row.row_content.display(row_data)
        if expanded:
            self._expandable_row.set_expanded(True)

    def _on_expand(self) -> None:
        self._add_location_rows(self._expanded_groups)
        self._add_secure_core_row(expanded=SecureCoreRow.LABEL.lower() in self._expanded_groups)

    def _on_collapse(self) -> None:
        self._remove_location_rows()
        self._remove_secure_core_row()

    def reset(self, keep_location_rows: bool = False):
        """Resets the country row to its initial state."""
        self._expandable_row.reset(keep_children=keep_location_rows)

    @property
    def country_name(self):
        """Returns this row's country name."""
        return self._country.name

    @property
    def country_code(self):
        """Returns this row's country code"""
        return self._country.code

    @property
    def expanded(self):
        """Returns whether the row is currently expanded or not."""
        return self._expandable_row.row_content.expanded

    @expanded.setter
    def expanded(self, value: bool):
        """Expands or collapses the country row (to show/hide its locations)."""
        self._expandable_row.row_content.expanded = value

    def grab_focus(self):  # pylint: disable=arguments-differ
        """See Gtk.Widget.grab_focus()"""
        self._expandable_row.row_content.grab_focus()

    def focus_on_location(self, location_name: str):
        """Focuses on the location in the country."""
        if not self.expanded:
            self.click_toggle_button()

        for location_row in self.location_rows:
            if location_row.label.lower() == location_name.lower():
                location_row.grab_focus()
                return

    @property
    def locations(self) -> List[Location]:
        """Returns the list of locations in the country."""
        runtime_assert(self._country is not None, "Country is not set")
        return self._country.locations

    @property
    def location_rows(self) -> List[LocationRow]:
        """Returns the list of location rows currently displayed."""
        return list(self._location_rows)

    @property
    def secure_core_row(self) -> Optional[SecureCoreRow]:
        """Returns the secure core row currently displayed."""
        return self._secure_core_row

    def click_toggle_button(self):
        """Simulates a click on the toggle button to expand/collapse the row."""
        self._expandable_row.row_content.click_toggle_button()

    def _remove_location_rows(self):
        while self._location_rows:
            location_row = self._location_rows.pop()
            self._location_row_container.remove(location_row)
            location_row.reset()

    def _remove_secure_core_row(self):
        """Removes the secure core row from its container."""
        if self._secure_core_row is not None:
            self._secure_core_row_container.remove(self._secure_core_row)
            self._secure_core_row.reset(keep_children=False)
            self._secure_core_row = None

    def _add_secure_core_row(self, expanded: bool = False):
        """Adds the single Via Secure Core row when the country has secure core servers."""
        runtime_assert(self._country is not None, "Country is not set")

        if not self._country.secure_core_group:
            self._remove_secure_core_row()
            return

        if self._secure_core_row is None:
            self._secure_core_row = SecureCoreRow()
            self._secure_core_row_container.append(self._secure_core_row)

        self._secure_core_row.display(
            self._controller, self._country.secure_core_group, self._user_tier,
            expanded=expanded
        )

    def _add_location_rows(self, expanded_locations: Optional[set[str]] = None):
        """Adds location rows to the country row.

        Args:
            expanded_locations: Optional set of lowercase location names that should be expanded
        """
        expanded_locations = expanded_locations or set()
        runtime_assert(self._country is not None, "Country is not set")

        locations = self._country.locations
        if self._user_tier == TierEnum.FREE and self._country.free:
            # If the current user has a free account, display first the free locations
            locations = list(chain(self._country.free_locations, self._country.paid_locations))

        def display_location_row(location_row, location):
            location_expanded = location.name.lower() in expanded_locations
            location_row.display(
                self._controller, location, self._user_tier, expanded=location_expanded
            )

        sync_rows_with_model_items(
            locations,
            self._location_rows,
            self._location_row_container,
            LocationRow,
            display_location_row
        )



"""
Generic exception handler.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import sys
import threading
from typing import TYPE_CHECKING, Optional
import gi

from proton.vpn.app.gtk.widgets.main.notifications import DialogButton
from proton.vpn.connection.exceptions import AuthenticationError, \
    HardJailedTwoFAError, NotYetValidCertificateError
from proton.session.exceptions import ProtonAPINotReachable, ProtonAPIError, \
    ProtonAPIAuthenticationNeeded, ProtonAPIMissingScopeError
from proton.vpn.session.exceptions import ServerNotFoundError
from proton.vpn import logging

gi.require_version("Gtk", "4.0")
from gi.repository import GLib, Gtk, Gio  # noqa: E402,E501 # pylint: disable=wrong-import-position,wrong-import-order

NO_SPACE_LEFT_ON_DEVICE_ERRNO = 28

if TYPE_CHECKING:
    from proton.vpn.app.gtk.controller import Controller
    from proton.vpn.app.gtk.widgets.main.main_widget import MainWidget

logger = logging.getLogger(__name__)


class ExceptionHandler:
    """Handles generic exceptions before they bubble all the way up."""
    GENERIC_ERROR_TITLE = "Something went wrong"
    GENERIC_ERROR_MESSAGE = "We're sorry, an unexpected error occurred." \
                            "Please try again."
    SERVER_NOT_FOUND_TITLE = "Unable to find server"
    PROTON_API_NOT_REACHABLE_MESSAGE = "Our servers are not reachable. " \
                                       "Please check your internet connection."
    VPN_AUTHENTICATION_ERROR_TITLE = "VPN connection error"
    VPN_AUTHENTICATION_ERROR_MESSAGE = (
        "Proton VPN could not connect to the VPN and blocked access to Internet to protect your IP."
        "\n\nClick \"Cancel Connection\" to restore your Internet connection. "
        "If the issue persists please try to sign out and in."
    )
    VPN_HARD_JAILED_2FA_ERROR_TITLE = "2FA Required"
    VPN_HARD_JAILED_2FA_ERROR_MESSAGE = (
        "You are connected to the VPN, but all traffic is blocked.\nYou need to"
        " go to the authentication page provided by security and authenticate"
        " with your hardware key.\nAfter that, the traffic will be enabled."
    )
    TIME_OUT_OF_SYNC_ERROR_TITLE = "Update system clock"
    TIME_OUT_OF_SYNC_ERROR_MESSAGE = (
        "Looks like your system clock is out of sync.\n"
        "This may cause issues when connecting to VPN.\n"
        "Update your system time and try to connect again."
    )

    def __init__(
        self, main_widget: Optional["MainWidget"] = None, controller: Optional["Controller"] = None
    ):
        super().__init__()
        self.main_widget = main_widget
        self.controller = controller
        self._previous_sys_excepthook = sys.excepthook
        self._previous_threading_excepthook = threading.excepthook

    def enable(self):
        """
        Enables the exception handler. Note that the exception handler
        should be enabled only after the main application window has been
        presented to the user so that error dialogs can actually be shown.
        """
        self._previous_sys_excepthook = sys.excepthook
        self._previous_threading_excepthook = threading.excepthook

        # Handle exceptions bubbling up in the main thread.
        sys.excepthook = self.handle_exception

        # Handle exceptions bubbling up in threads started with Thread.run().
        # Notice that an exception raised from a thread managed by a
        # ThreadPoolExecutor won't bubble up, as the executor won't allow it.
        # In this case, make sure that you call Future.result() on the future
        # returned by ThreadPoolExecutor.submit() in the main thread
        # (e.g. using GLib.idle_add()).
        threading.excepthook = self.handle_thread_exception

    def disable(self):
        """Disables the exception handler. The exception handler should
        be disabled as soon as the main application window is closed. This
        is specially important in tests, as the python process might run
        other test code."""
        sys.excepthook = self._previous_sys_excepthook
        threading.excepthook = self._previous_threading_excepthook

    def handle_thread_exception(self, args):
        """
        When the application exception handler is enabled, this method
        is triggered on errors that happened on threads.
        :param args: dictionary passed to threading.excepthook. For more info:
        https://docs.python.org/3/library/threading.html#threading.excepthook
        """
        GLib.idle_add(
            self.handle_exception,
            args.exc_type, args.exc_value, args.exc_traceback
        )

    def handle_exception(self, exc_type, exc_value, exc_traceback):
        """
        When the application exception handler is enabled, this method is
        triggered on errors that were not explicitly handled by the main
        application thread.
        :param exc_type: Type of the exception.
        :param exc_value: Instance of the exception. It might be None if the
        exception was triggered using an Exception class, rather than an object
        (e.g. raise Exception, instead of raise Exception()).
        :param exc_traceback: The exception traceback.
        """
        if issubclass(exc_type, ProtonAPIAuthenticationNeeded):
            self._on_proton_api_authentication_needed()
        elif issubclass(exc_type, ProtonAPINotReachable):
            self._on_proton_api_not_reachable(exc_type, exc_value, exc_traceback)
        elif isinstance(exc_value, ProtonAPIMissingScopeError):
            self._on_proton_api_missing_scope_error(exc_value)
        elif isinstance(exc_value, ProtonAPIError) and exc_value.error:
            self._on_proton_api_error(exc_type, exc_value, exc_traceback)
        elif isinstance(exc_value, ServerNotFoundError):
            self._on_server_not_found(exc_type, exc_value, exc_traceback)
        elif isinstance(exc_value, AuthenticationError):
            self._on_vpn_authentication_error(exc_type, exc_value, exc_traceback)
        elif isinstance(exc_value, HardJailedTwoFAError):
            self._on_hard_jailed_2fa_error(exc_type, exc_value, exc_traceback)
        elif isinstance(exc_value, NotYetValidCertificateError):
            self._on_certificate_not_yet_valid_error(exc_type, exc_value, exc_traceback)
        elif isinstance(exc_value, OSError) and exc_value.errno == NO_SPACE_LEFT_ON_DEVICE_ERRNO:
            self._on_no_space_left_on_device(exc_type, exc_value, exc_traceback)
        elif issubclass(exc_type, AssertionError):
            # We shouldn't catch assertion errors raised by tests.
            raise exc_value
        elif issubclass(exc_type, Exception):
            self._on_exception(exc_type, exc_value, exc_traceback)
        else:
            raise exc_value if exc_value else exc_type

    def _on_proton_api_authentication_needed(self):
        logger.warning(
            "Authentication required.",
            category="API", event="WARNING"
        )
        if self.main_widget:
            self.main_widget.on_session_expired()

    def _on_proton_api_not_reachable(self, exc_type, exc_value, exc_traceback):
        if self.main_widget:
            self.main_widget.notifications.show_error_message(
                self.PROTON_API_NOT_REACHABLE_MESSAGE,
            )
        logger.warning(
            "API not reachable.",
            category="API", event="ERROR",
            exc_info=(exc_type, exc_value, exc_traceback)
        )

    def _on_proton_api_missing_scope_error(self, missing_scope_error: ProtonAPIMissingScopeError):
        """The user is valid but lacks VPN permissions."""
        logger.warning(
            missing_scope_error.error,
            category="APP", event="WARNING"
        )

        if self.main_widget:
            self._logout_and_show_missing_scope_dialog(missing_scope_error)

    def _logout_and_show_missing_scope_dialog(self, error: ProtonAPIMissingScopeError):
        """This method is called by the exception handler when the user
        lacks VPN permissions."""
        self.main_widget.logout()

        def on_dialog_closed(response_type: Gtk.ResponseType):
            if Gtk.ResponseType.OK == response_type:
                Gio.AppInfo.launch_default_for_uri(
                    "https://protonvpn.com/support/assign-vpn-connection", None
                )

        error_details = error.json_data["Details"]
        actions = error_details.get("Actions", [])

        buttons = []
        for action in actions:
            if action["Code"] == "AssignConnections":
                buttons.append(
                    DialogButton(label=action["Name"], response_type=Gtk.ResponseType.OK)
                )
        buttons.append(
            DialogButton(label="Sign in again", response_type=Gtk.ResponseType.CLOSE)
        )

        self.main_widget.notifications.show_error_dialog(
            title=error_details["Title"],
            message=error_details["Body"],
            hint=error_details.get("Hint"),
            message_type=Gtk.MessageType.OTHER,
            buttons=buttons,
            on_dialog_closed=on_dialog_closed
        )

    def _on_proton_api_error(self, exc_type, exc_value, exc_traceback):
        if self.main_widget:
            self.main_widget.notifications.show_error_message(exc_value.error)
        logger.error(
            exc_value.error,
            category="APP", event="ERROR",
            exc_info=(exc_type, exc_value, exc_traceback)
        )

    def _on_server_not_found(self, exc_type, exc_value, exc_traceback):
        if self.main_widget:
            self.main_widget.notifications.show_error_dialog(
                title=self.SERVER_NOT_FOUND_TITLE,
                message=str(exc_value)
            )
        logger.error(
            exc_value,
            category="APP", event="ERROR",
            exc_info=(exc_type, exc_value, exc_traceback)
        )

    def _on_vpn_authentication_error(self, exc_type, exc_value, exc_traceback):
        if self.main_widget:
            self.main_widget.notifications.show_error_dialog(
                title=self.VPN_AUTHENTICATION_ERROR_TITLE,
                message=self.VPN_AUTHENTICATION_ERROR_MESSAGE
            )

        logger.error(
            exc_value,
            category="APP", event="ERROR",
            exc_info=(exc_type, exc_value, exc_traceback)
        )

    def _on_hard_jailed_2fa_error(self, exc_type, exc_value, exc_traceback):
        if self.main_widget:
            self.main_widget.notifications.show_error_dialog(
                title=self.VPN_HARD_JAILED_2FA_ERROR_TITLE,
                message=self.VPN_HARD_JAILED_2FA_ERROR_MESSAGE
            )

        logger.error(
            exc_value,
            category="APP", event="ERROR",
            exc_info=(exc_type, exc_value, exc_traceback)
        )

    def _on_certificate_not_yet_valid_error(self, exc_type, exc_value, exc_traceback):
        if self.main_widget:
            self.main_widget.notifications.show_error_dialog(
                title=self.TIME_OUT_OF_SYNC_ERROR_TITLE,
                message=self.TIME_OUT_OF_SYNC_ERROR_MESSAGE
            )

        logger.error(
            exc_value,
            category="APP", event="ERROR",
            exc_info=(exc_type, exc_value, exc_traceback)
        )

    def _on_no_space_left_on_device(self, exc_type, exc_value, exc_traceback):
        logger.error(
            exc_value,
            category="APP", event="ERROR",
            exc_info=(exc_type, exc_value, exc_traceback)
        )
        if self.main_widget:
            self.main_widget.notifications.show_error_dialog(
                title="No space left on device",
                message="There is not enough space left on your device."
            )

    def _on_exception(self, exc_type, exc_value, exc_traceback):
        if self.main_widget:
            self.main_widget.notifications.show_error_dialog(
                title=self.GENERIC_ERROR_TITLE,
                message=self.GENERIC_ERROR_MESSAGE,
            )
        logger.critical(
            "Unexpected error.",
            category="APP", event="CRASH",
            exc_info=(exc_type, exc_value, exc_traceback)
        )

        if self.controller:
            self.controller.send_error_to_proton(
                (exc_type, exc_value, exc_traceback))

    def __enter__(self):
        self.enable()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disable()



"""
Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.


Memory-safe GObject signal connections.

Connects callbacks to GObject signals while statically rejecting any
use that could leak. Two callback types are valid:

1. Bound methods (``self.on_clicked``). The bound instance is held
   weakly via ``weakref.WeakMethod``, so the signal connection alone
   cannot keep the bound instance alive.

2. Pure free functions (``__closure__`` is empty). They reference
   nothing about any instance — no leak is possible. They are
   connected directly without WeakMethod.

Anything else is rejected at connection time:

- Closure-capturing callables (lambdas, nested functions with captures,
  ``functools.partial`` over a bound method, etc.) — closure cells hold
  their captures strongly and may leak whatever they reference. Raises
  ``TypeError``.
- The ``"destroy"`` signal is rejected explicitly. By the time destroy
  fires from inside ``pygobject_dealloc`` the wrapper's weakrefs have
  already been cleared and the handler is silently skipped. Override
  ``do_dispose`` instead. Raises ``ValueError``.

The signature is fixed at three positional arguments — extra
``user_data`` forwarded to the handler (the GObject ``connect`` idiom)
is not supported. If you need to dispatch extra context, store it on
``self`` and read it inside the bound method.

A keyword-only ``once=True`` flag connects a one-shot handler that
self-disconnects from *emitter* after firing once, or after the bound
instance is collected — whichever comes first. The same callback
restrictions apply.

Usage::

    safe_signal_connect(button, "clicked", self.on_clicked)              # bound method
    safe_signal_connect(button, "clicked", log_click)                    # pure free function
    safe_signal_connect(button, "clicked", self.on_clicked, once=True)   # one-shot
"""
import inspect
import types
import weakref
from typing import Callable

from gi.repository import GObject

from proton.vpn import logging

logger = logging.getLogger(__name__)


def safe_signal_connect(
    emitter: GObject.Object,
    signal: str,
    callback: Callable,
    *,
    once=False
) -> int:
    """Connect *callback* to *emitter*'s *signal*.

    Bound methods are held weakly via ``WeakMethod``; pure free functions
    (no closure captures) are connected directly. Closure-capturing
    callables raise ``TypeError``. The ``"destroy"`` signal raises
    ``ValueError`` (override ``do_dispose`` instead).

    If *once* is True, the handler self-disconnects from *emitter* after
    firing once (or after the bound instance is collected, whichever
    comes first).

    Returns the GObject handler id so the caller can disconnect manually
    if needed."""
    if signal == "destroy":
        raise ValueError(
            "safe_signal_connect cannot wrap the 'destroy' signal: by the "
            "time destroy fires from inside pygobject_dealloc, the wrapper's "
            "weakrefs have already been cleared and WeakMethod() returns "
            "None, so the handler is silently skipped. Override do_dispose "
            "instead:\n"
            "    def do_dispose(self):\n"
            "        # ...your cleanup here, idempotent...\n"
            "        SuperClass.do_dispose(self)  # e.g. Gtk.Box.do_dispose"
        )

    if inspect.ismethod(callback):
        return _connect_weak(emitter, signal, callback, once=once)

    if not callable(callback):
        raise TypeError(
            f"safe_signal_connect callback must be callable, got "
            f"{type(callback).__name__}."
        )

    if _has_closure_captures(callback):
        raise TypeError(
            "safe_signal_connect rejects closure-capturing callbacks "
            "(lambdas, nested functions with captures): closure cells hold "
            "their captures strongly and may leak. Pass a bound method, or "
            "a pure free function with no closure captures."
        )

    return _connect(emitter, signal, callback, once=once)


def _has_closure_captures(callback: Callable) -> bool:
    """True if *callback* is a function-like with non-empty closure cells."""
    return bool(getattr(callback, "__closure__", None))


def _connect_weak(
    emitter: GObject.Object,
    signal: str,
    method: types.MethodType,
    *,
    once: bool = False,
) -> int:
    """Connect via ``weakref.WeakMethod`` so the handler does not extend
    the bound instance's lifetime. The wrapper transparently dispatches
    to the live bound method; once the instance is collected, the wrapper
    no-ops and the slot stays inert in *emitter*'s signal table until
    *emitter* itself is destroyed.

    If ``once=True``, the wrapper self-disconnects from *emitter* after
    firing once."""
    weak_method = weakref.WeakMethod(method)
    weak_emitter = weakref.ref(emitter)
    handler_id = None

    # Captured as plain strings so the wrapper closure holds no strong
    # reference to emitter or the bound instance.
    receiver_repr = (
        f"{type(method.__self__).__name__}.{method.__func__.__name__}"
    )
    emitter_repr = f"{type(emitter).__name__}@{hex(id(emitter))}"

    def wrapper(*args):
        if once and handler_id is not None:
            live_emitter = weak_emitter()
            if live_emitter is not None:
                live_emitter.disconnect(handler_id)

        target = weak_method()
        if target is not None:
            return target(*args)

        logger.warning(
            "Signal smell: '%s' fired on %s but receiver %s was already "
            "collected — handler no-op'd. The signal connection outlived its listener",
            signal, emitter_repr, receiver_repr,
        )
        return None

    handler_id = emitter.connect(signal, wrapper)
    return handler_id


def _connect(
    emitter: GObject.Object,
    signal: str,
    callback: Callable,
    *,
    once: bool = False,
) -> int:
    """Connect a pure free function (no closure captures). For persistent
    connections this delegates directly to ``emitter.connect``. With
    ``once=True``, a small self-disconnecting wrapper is used."""
    if not once:
        return emitter.connect(signal, callback)

    weak_emitter = weakref.ref(emitter)
    handler_id = None

    def wrapper(*args):
        if handler_id is not None:
            live_emitter = weak_emitter()
            if live_emitter is not None:
                live_emitter.disconnect(handler_id)

        callback(*args)

    handler_id = emitter.connect(signal, wrapper)
    return handler_id
